"""
Workflow from computing the band structures
"""
from copy import deepcopy

import aiida.orm as orm
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import WorkChain, calcfunction, if_
from aiida.plugins import WorkflowFactory

from ..common import OUTPUT_LINKNAMES as out_ln
from ..common import INPUT_LINKNAMES as inp_ln


class CastepBandsWorkChain(WorkChain):
    """
    Workchain for running bands calculation.

    This workchain does the following:

    1. Relax the structure if requested (eg. inputs passed to the relax namespace).
    2. Optionally: Do a SCF singlepoint calculation
    3. Do combined SCF + non-SCF calculation for bands and dos.

    Inputs must be passed for the SCF calculation (dispatched to bands and DOS), 
    others are optional.

    Input for bands and dos calculations are optional. However, if they are needed, the full list of inputs must
    be passed. For the `parameters` node, one may choose to only specify those fields that need to be updated. 
    """
    _base_wk_string = 'castep.base'
    _relax_wk_string = 'castep.relax'
    _task_name = 'spectral'

    @classmethod
    def define(cls, spec):
        """Initialise the WorkChain class"""
        super().define(spec)
        relax_work = WorkflowFactory(cls._relax_wk_string)
        base_work = WorkflowFactory(cls._base_wk_string)

        spec.input('structure',
                   help='The input structure',
                   valid_type=orm.StructureData)
        spec.input('bands_kpoints',
                   help='Explicit kpoints for the bands',
                   valid_type=orm.KpointsData,
                   required=False)
        spec.input('bands_kpoints_distance',
                   help='Spacing for band distances, used by seekpath',
                   valid_type=orm.Float,
                   required=False)
        spec.input(
            'dos_kpoints',
            help='Kpoints for running DOS calculations',
            required=False,
            valid_type=orm.KpointsData,
        )
        spec.expose_inputs(relax_work,
                           namespace='relax',
                           exclude=('structure', ),
                           namespace_options={
                               'required': False,
                               'populate_defaults': False,
                               'help':
                               'Inputs for Relaxation workchain, if needed'
                           })
        spec.expose_inputs(
            base_work,
            namespace='scf',
            exclude=('calc.structure', 'calc.kpoints'),
            namespace_options={
                'required':
                True,
                'populate_defaults':
                True,
                'help':
                'Inputs for SCF workchain, mandatory. Explicit kpoints is not allowed.'
            })
        spec.expose_inputs(base_work,
                           namespace='bands',
                           exclude=('calc.structure', 'calc.kpoints'),
                           namespace_options={
                               'required': False,
                               'populate_defaults': False,
                               'help':
                               'Inputs for bands calculation, if needed'
                           })
        spec.expose_inputs(base_work,
                           namespace='dos',
                           exclude=('calc.structure', 'calc.kpoints'),
                           namespace_options={
                               'required': False,
                               'populate_defaults': False,
                               'help': 'Inputs for DOS calculation, if needed'
                           })
        spec.input('clean_children_workdir',
                   valid_type=orm.Str,
                   help='What part of the called children to clean',
                   required=False,
                   default=lambda: orm.Str('none'))
        spec.input('only_dos',
                   required=False,
                   help='Flag for running only DOS calculations')
        spec.input(
            'run_separate_scf',
            required=False,
            help='Flag for running a separate SCF calculation, default to False'
        )
        spec.outline(
            cls.setup,
            if_(cls.should_do_relax)(
                cls.run_relax,
                cls.verify_relax,
            ),
            if_(cls.should_run_seekpath)(cls.run_seekpath),
            if_(cls.should_run_scf)(
                cls.run_scf,
                cls.verify_scf,
            ),
            cls.run_bands_dos,
            cls.inspect_bands_dos,
        )

        spec.output(
            'primitive_structure',
            help='Primitive structure used for band structure calculations')
        spec.output('band_structure',
                    help='Computed band structure with labels')
        spec.output('seekpath_parameters',
                    help='Parameters used by seekpath',
                    required=False)
        spec.output('dos_bands',
                    required=False,
                    help='Bands from the DOS calculation')

        spec.exit_code(501,
                       'ERROR_SUB_PROC_RELAX_FAILED',
                       message='Relaxation workchain failed')
        spec.exit_code(502,
                       'ERROR_SUB_PROC_SCF_FAILED',
                       message='SCF workchain failed')
        spec.exit_code(503,
                       'ERROR_SUB_PROC_BANDS_FAILED',
                       message='Band structure workchain failed')
        spec.exit_code(504,
                       'ERROR_SUB_PROC_DOS_FAILED',
                       message='DOS workchain failed')

    def setup(self):
        """Setup the calculation"""
        self.ctx.current_structure = self.inputs.structure
        self.ctx.bands_kpoints = self.inputs.get('bands_kpoints')
        if 'kpoints' in self.inputs.scf:
            self.report(
                "WARNING: explicit kpoints for SCF - this is likely to cause problems"
            )

    def should_do_relax(self):
        """Wether we should do relax or not"""
        return 'relax' in self.inputs

    def run_relax(self):
        """Run the relaxation"""
        relax_work = WorkflowFactory(self._relax_wk_string)
        inputs = self.exposed_inputs(relax_work, 'relax', agglomerate=True)
        inputs = AttributeDict(inputs)
        inputs.metadata.call_link_label = 'relax'
        inputs.structure = self.ctx.current_structure

        running = self.submit(relax_work, **inputs)
        return self.to_context(workchain_relax=running)

    def verify_relax(self):
        """Verify the relaxation"""
        relax_workchain = self.ctx.workchain_relax
        if not relax_workchain.is_finished_ok:
            self.report('Relaxation finished with Error')
            return self.exit_codes.ERROR_SUB_PROC_RELAX_FAILED

        # Use the relaxed structure as the current structure
        self.ctx.current_structure = relax_workchain.outputs[
            out_ln['structure']]

    def should_run_scf(self):
        """Wether we should run SCF calculation?"""
        run_separate_scf = self.inputs.get('run_separate_scf', orm.Bool(False))
        return run_separate_scf.value

    def should_run_seekpath(self):
        """Seekpath should only run if no explicit bands is provided"""
        return 'bands_kpoints' not in self.inputs

    def run_seekpath(self):
        """
        Run seekpath to obtain the primitive structure and bands
        """

        inputs = {
            'reference_distance': self.inputs.get('bands_kpoints_distance',
                                                  None),
            'metadata': {
                'call_link_label': 'seekpath'
            }
        }

        results = seekpath_structure_analysis(self.ctx.current_structure,
                                              **inputs)
        self.ctx.current_structure = results['primitive_structure']
        self.ctx.bands_kpoints = results['explicit_kpoints']
        self.out('primitive_structure', results['primitive_structure'])
        self.out('seekpath_parameters', results['parameters'])

    def run_scf(self):
        """
        Run the SCF calculation
        """

        base_work = WorkflowFactory(self._base_wk_string)
        inputs = AttributeDict(self.exposed_inputs(base_work, namespace='scf'))
        inputs.metadata.call_link_label = 'scf'
        inputs.calc.structure = self.ctx.current_structure

        # Ensure that writing the check/castep_bin
        param_dict = inputs.calc.parameters.get_dict()
        if 'PARAM' in param_dict:
            ensure_checkpoint(param_dict['PARAM'])
        else:
            ensure_checkpoint(param_dict)

        # Update if changes are made
        if param_dict != inputs.calc.parameters.get_dict():
            self.report(
                "Updated the PARAM to make sure castep_bin file will be written"
            )
            inputs.calc.parameters = orm.Dict(dict=param_dict)

        running = self.submit(base_work, **inputs)
        self.report('Running SCF calculation {}'.format(running))
        self.to_context(workchain_scf=running)

    def verify_scf(self):
        """Inspect the SCF calculation"""
        scf_workchain = self.ctx.workchain_scf
        if not scf_workchain.is_finished_ok:
            self.report('SCF workchain finished with Error')
            return self.exit_codes.ERROR_SUB_PROC_SCF_FAILED

        # NOTE: the plugin does not support restarting from local files for now,
        # This should be added later - restart from a `retrieved` local folder
        self.ctx.restart_folder = scf_workchain.outputs.remote_folder
        self.report("SCF calculation {} completed".format(scf_workchain))

    def run_bands_dos(self):
        """Run the bands and the DOS calculations"""
        base_work = WorkflowFactory(self._base_wk_string)
        # Use the SCF inputs as the base
        inputs = AttributeDict(self.exposed_inputs(base_work, namespace='scf'))
        inputs.calc.structure = self.ctx.current_structure
        if self.ctx.get('restart_folder'):
            inputs.continuation_folder = self.ctx.restart_folder
            has_scf = True
        else:
            has_scf = False

        running = {}
        only_dos = self.inputs.get('only_dos')

        def generate_sub_input(inputs, namespace, task):
            """
            Generate inputs for tasks, merge those in the namespace from those
            given in the inputs
            """
            if namespace in self.inputs:
                self.report(
                    'Taking input from the {} namespace'.format(namespace))
                bands_inputs = AttributeDict(
                    self.exposed_inputs(base_work, namespace=namespace))
            else:
                bands_inputs = AttributeDict(
                    {'calc': {
                        'parameters': orm.Dict(dict={'task': task})
                    }})

            # Special treatment - combine the paramaters
            parameters = inputs.calc.parameters.get_dict()
            bands_parameters = bands_inputs.calc.parameters.get_dict()

            nested_update(parameters, bands_parameters)
            # Make sure the task name is correct
            nested_update(parameters, {'task': self._task_name})

            # Update the SCF name space with those from the bands name space
            nested_update(inputs, bands_inputs)

            # Apply the new parameters
            inputs.calc.parameters = orm.Dict(dict=parameters)

            return inputs

        if (only_dos is None) or (only_dos.value is False):
            inputs = generate_sub_input(inputs, 'bands', 'spectral')
            # Set the kpoints
            inputs.calc[self._task_name + '_kpoints'] = self.ctx.bands_kpoints
            bands_calc = self.submit(base_work, **inputs)
            running['bands_workchain'] = bands_calc
            self.report(
                'Submitted workchain {} for band structure'.format(bands_calc))

        if ('dos_kpoints' in self.inputs) or ('dos' in self.inputs):

            inputs = generate_sub_input(inputs, 'dos', 'spectral')
            # Set the kpoints
            inputs.calc[self._task_name + '_kpoints'] = self.inputs.dos_kpoints
            dos_calc = self.submit(base_work, **inputs)
            running['dos_workchain'] = dos_calc
            self.report(
                'Submitted workchain {} for dos calculation'.format(dos_calc))
        return self.to_context(**running)

    def inspect_bands_dos(self):
        """Inspect the bands and dos calculations"""

        exit_code = None

        if 'bands_workchain' in self.ctx:
            bands = self.ctx.bands_workchain
            if not bands.is_finished_ok:
                self.report(
                    'Bands calculation finished with error, exit_status: {}'.
                    format(bands, bands.exit_status))
                exit_code = self.exit_codes.ERROR_SUB_PROC_BANDS_FAILED
            self.out(
                'band_structure',
                compose_labelled_bands(bands.outputs[out_ln['bands']],
                                       self.ctx.bands_kpoints))
        else:
            bands = None

        if 'dos_workchain' in self.ctx:
            dos = self.ctx.dos_workchain
            if not dos.is_finished_ok:
                self.report(
                    'DOS calculation finished with error, exit_status: {}'.
                    format(dos.exit_status))
                exit_code = self.exit_codes.ERROR_SUB_PROC_DOS_FAILED
            self.out('dos_bands', dos.outputs[out_ln['bands']])

        else:
            dos = None

        return exit_code

    def on_terminated(self):
        """
        Clean the remote directories of all called childrens
        """

        super(CastepBandsWorkChain, self).on_terminated()

        if self.inputs.clean_children_workdir.value != 'none':
            cleaned_calcs = []
            for called_descendant in self.node.called_descendants:
                if isinstance(called_descendant, orm.CalcJobNode):
                    try:
                        called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                        cleaned_calcs.append(called_descendant.pk)
                    except (IOError, OSError, KeyError):
                        pass

            if cleaned_calcs:
                self.report(
                    'cleaned remote folders of calculations: {}'.format(
                        ' '.join(map(str, cleaned_calcs))))


def nested_update(dict_in, update_dict):
    """Update the dictionary - combine nested subdictionary with update as well"""
    for key, value in update_dict.items():
        if key in dict_in and isinstance(value, (dict, AttributeDict)):
            nested_update(dict_in[key], value)
        else:
            dict_in[key] = value
    return dict_in


def nested_update_dict_node(dict_node, update_dict):
    """Utility to update a Dict node in a nested way"""
    pydict = dict_node.get_dict()
    nested_update(pydict, update_dict)
    if pydict == dict_node.get_dict():
        return dict_node
    else:
        return orm.Dict(dict=pydict)


@calcfunction
def seekpath_structure_analysis(structure, **kwargs):
    """Primitivize the structure with SeeKpath and generate the high symmetry k-point path through its Brillouin zone.
    This calcfunction will take a structure and pass it through SeeKpath to get the normalized primitive cell and the
    path of high symmetry k-points through its Brillouin zone. Note that the returned primitive cell may differ from the
    original structure in which case the k-points are only congruent with the primitive cell.
    The keyword arguments can be used to specify various Seekpath parameters, such as:
        with_time_reversal: True
        reference_distance: 0.025
        recipe: 'hpkot'
        threshold: 1e-07
        symprec: 1e-05
        angle_tolerance: -1.0
    Note that exact parameters that are available and their defaults will depend on your Seekpath version.
    """
    from aiida.tools import get_explicit_kpoints_path

    unwrapped_kwargs = {
        key: node.value
        for key, node in kwargs.items() if isinstance(node, orm.Data)
    }

    # All keyword arugments should be `Data` node instances of base type and so should have the `.value` attribute
    return get_explicit_kpoints_path(structure, **unwrapped_kwargs)


@calcfunction
def compose_labelled_bands(bands, kpoints):
    """
    Add additional information from the kpoints allow richer informations
    to be stored such as band structure labels.
    """
    new_bands = deepcopy(bands)
    new_bands.set_kpointsdata(kpoints)
    return new_bands


def ensure_checkpoint(pdict):
    """Ensure that check/castep_bin file will be wirtten"""
    value = pdict.get('write_checkpoint')
    if value is None:
        return pdict
    elif value.lower() == 'none':
        pdict['write_checkpoint'] = 'minimal'
    return pdict



"""
Utilities module
This module does not have top level AiiDA orm imports
"""
from __future__ import absolute_import
import re
import os


def split_otfg_entry(otfg):
    """
    Split an entry of otfg in the form of element_settings
    :returns (element, entry):
    """
    otfg = otfg.strip()
    # Check it is a library
    lib = re.match(r'^[A-Za-z]+\d+$', otfg)
    if lib:
        element = "LIBRARY"
        return element, otfg

    # It not a library, try to get the element
    match = re.match(r"^([A-Z][a-z]?)[ _]+(.+)$", otfg)
    if match:
        element, setting = match.group(1), match.group(2)
    else:
        element = None
        setting = otfg

    return element, setting


def get_usp_element(filepath):
    """
    infer element from usp/recpot filename
    :return element: a string of element name or None if not found
    """
    # Convert to string -> allow Path object to be passed
    filepath = str(filepath)
    re_fn = re.compile(r"^([a-z]+)[_-].+\.(usp|recpot)$", flags=re.IGNORECASE)
    filename = os.path.split(filepath)[1]
    match = re_fn.match(filename)
    if match:
        element = match.group(1)
        element = element.capitalize()
        return element
    return None



"""
Use pytest 
"""
# pylint: disable=unused-import,unused-argument,redefined-outer-name,too-many-function-args,
# pylint: disable=protected-access,abstract-class-instantiated,no-value-for-parameter,unexpected-keyword-arg

from __future__ import absolute_import
import pytest
import six
from pathlib import Path
from aiida.common import ValidationError

Ti_otfg = "Ti 3|1.8|9|10|11|30U:40:31:32(qc=5.5)"
Sr_otfg = "Sr 3|2.0|5|6|7|40U:50:41:42"
O_otfg = "O 2|1.1|15|18|20|20:21(qc=7)"
raw_otfg = "2|1.1|15|18|20|20:21(qc=7)"
otfglibs = ["QC5", "NCP17", "C9"]

OTFG_COLLECTION = {"Sr": Sr_otfg, "Ti": Ti_otfg, "O": O_otfg}


@pytest.fixture(scope="module")
def otfgdata():
    from aiida_castep.data.otfg import OTFGData
    return OTFGData


@pytest.fixture(scope="module")
def otfg():
    import aiida_castep.data.otfg as otfg
    return otfg


@pytest.fixture(scope="module")
def imps(aiida_profile):
    class Imports(object):
        from aiida.plugins import DataFactory
        import aiida_castep.data.otfg as otfg

    return Imports


def test_otfg_create(new_database, imps, otfg, otfgdata):
    """
    Test for creating OTFGData notes
    """
    element, setting = otfg.split_otfg_entry(Ti_otfg)
    C9 = otfgdata(otfg_entry="C9")
    assert C9.string == "C9"
    assert C9.element == "LIBRARY"
    assert C9.entry == "C9"
    C9.store()

    ti = otfgdata(otfg_entry=setting)
    ti.set_element(element)
    assert ti.string == setting
    assert ti.element == element
    assert ti.entry == Ti_otfg

    ti.store()

    Sr = otfgdata()
    Sr.set_string(Sr_otfg)
    Sr.delete_attribute("element")
    with pytest.raises(ValidationError):
        Sr.store()


def test_otfg_get_or_create(new_database, otfg, otfgdata):
    Ti, create = otfgdata.get_or_create(Ti_otfg, store_otfg=False)
    assert create is True

    # Create but not stored do it again should have no change
    Ti, create = otfgdata.get_or_create(Ti_otfg, store_otfg=True)
    assert create is True

    # Now we are simply getting the Ti otfg data
    Ti2, create = otfgdata.get_or_create(Ti_otfg, store_otfg=False)
    assert create is False

    # Should get the stored OTFG entry
    # If using SQLA this Ti should be IS Ti2?
    assert Ti2.uuid == Ti.uuid

    # Using another way should be the same
    Sr = otfgdata()
    Sr.set_element("Sr")
    Sr.set_string("bla")
    Sr.store()

    Sr2, create = otfgdata.get_or_create("Sr bla")
    assert create is False

    # A different OTFG
    Sr3, create = otfgdata.get_or_create("Sr foo")
    assert create is True

    # Check if more than one instance is found in the db
    with pytest.raises(ValueError):
        Sr = otfgdata()
        Sr.set_element("Sr")
        Sr.set_string("bla")
        Sr.store()
        otfgdata.get_or_create("Sr bla", use_first=False)


@pytest.fixture
def otfg_nodes(aiida_profile, otfgdata):
    otfgs = {}
    for symbol in ["Sr", "Ti", "O"]:
        otemp = otfgdata()
        otemp.set_element(symbol)
        otemp.set_string("Foo")
        otfgs[symbol] = otemp

    return list(otfgs.values())


def test_set_up_family_from_string(new_database, imps, otfg_nodes, otfgdata):

    text_entries = [n.entry for n in otfg_nodes]
    entry, uploaded = imps.otfg.upload_otfg_family(text_entries[:1], "Test",
                                                   "Test")
    assert (entry, uploaded) == (1, 1)

    # Creating duplicated family - should fail
    with pytest.raises(ValidationError):
        entry, uploaded = imps.otfg.upload_otfg_family(text_entries, "Test",
                                                       "Test")

    entry, uploaded = imps.otfg.upload_otfg_family(text_entries,
                                                   "Test",
                                                   "Test",
                                                   stop_if_existing=False)

    assert (entry, uploaded) == (3, 2)

    # Check if they are indeed stored
    groups = otfgdata.get_otfg_groups()
    retrieved_entries = [node.entry for node in groups[0].nodes]

    for txt in text_entries:
        assert txt in retrieved_entries


def test_set_up_family_from_nodes(new_database, otfg, otfg_nodes, otfgdata):

    entry, uploaded = otfg.upload_otfg_family(otfg_nodes[:1],
                                              "Test",
                                              "Test",
                                              stop_if_existing=True)

    groups = otfgdata.get_otfg_groups()
    assert len(groups) == 1
    assert len(groups[0].nodes) == 1

    with pytest.raises(ValidationError):
        entry, uploaded = otfg.upload_otfg_family(otfg_nodes,
                                                  "Test",
                                                  "Test",
                                                  stop_if_existing=True)

    entry, uploaded = otfg.upload_otfg_family(otfg_nodes,
                                              "Test",
                                              "Test",
                                              stop_if_existing=False)
    groups = otfgdata.get_otfg_groups()
    assert len(groups) == 1
    assert len(groups[0].nodes) == 3

    Ce = otfgdata()
    Ce.set_string("Ce BLA")

    entry, uploaded = otfg.upload_otfg_family(otfg_nodes + [Ce],
                                              "Test",
                                              "Test",
                                              stop_if_existing=False)

    group = otfgdata.get_otfg_group("Test")

    uuid_in_the_group = [node.uuid for node in group.nodes]
    for o in otfg_nodes + [Ce]:
        assert o.uuid in uuid_in_the_group


def test_assign_from_structure(
        new_database,
        db_test_app,
):
    """
    Test using get_pseudos_from_structure
    """

    from aiida_castep.data import get_pseudos_from_structure
    from aiida.common import NotExistent
    from ..utils import get_sto_structure

    db_test_app.upload_otfg_family([Sr_otfg, Ti_otfg, O_otfg], "STO")

    STO = get_sto_structure()

    pseudo_list = get_pseudos_from_structure(STO, "STO")
    assert pseudo_list["Sr"].entry == OTFG_COLLECTION["Sr"]
    assert pseudo_list["O"].entry == OTFG_COLLECTION["O"]
    assert pseudo_list["Ti"].entry == OTFG_COLLECTION["Ti"]

    with pytest.raises(NotExistent):
        pseudo_list = get_pseudos_from_structure(STO, "STO_O_missing")

    db_test_app.upload_otfg_family([Sr_otfg, Ti_otfg, O_otfg, "C9"],
                                   "STO+C9",
                                   stop_if_existing=False)
    STO.append_atom(symbols=["Ce"], position=[0, 0, 0])
    pseudo_list = get_pseudos_from_structure(STO, "STO+C9")
    assert pseudo_list["Sr"].entry == OTFG_COLLECTION["Sr"]
    assert pseudo_list["O"].entry == OTFG_COLLECTION["O"]
    assert pseudo_list["Ti"].entry == OTFG_COLLECTION["Ti"]
    assert pseudo_list["Ce"].entry == "C9"


@pytest.fixture
def usp_folder(new_workdir):
    import os
    for fn in ["Sr_00.usp", "Ti-00.usp", "Ce_00.recpot"]:
        with open(str(new_workdir / fn), "w") as fh:
            fh.write("Bla " + fn)
    return new_workdir


def test_usp_upload_family(new_database, usp_folder):
    """
    Test uploading the usp family
    """
    from aiida_castep.data.usp import upload_usp_family
    from aiida_castep.data.otfg import upload_otfg_family, OTFGGroup
    upload_usp_family(str(usp_folder), "Test", "Test")

    new = usp_folder / "O_00.usp"
    new.write_text(six.text_type("asdfgghhd"))
    # This will raise an exception as the same file is being uploaded
    with pytest.raises(ValueError):
        upload_usp_family(str(usp_folder),
                          "Test",
                          "Test",
                          stop_if_existing=True)
    # This should be OK
    upload_usp_family(str(usp_folder), "Test", "Test", stop_if_existing=False)

    # Should be able to mix-in Usp and OTFG entries
    upload_otfg_family(['C18'], "Test", "Test")

    nodes = list(OTFGGroup.get(label='Test').nodes)
    upload_otfg_family(nodes, "Test2", "Test", stop_if_existing=False)


def test_usp_get_or_create(new_database, usp_folder):
    """Testing the logic or get_or_create"""
    import aiida_castep.data.usp as usp
    fpath = usp_folder / "Sr_00.usp"
    node1, create = usp.UspData.get_or_create(fpath)

    assert create is True
    assert node1.element == "Sr"

    node2 = usp.UspData(file=fpath)
    node2.store()

    # Now having two files - should raise an exception
    with pytest.raises(ValueError):
        node3, create = usp.UspData.get_or_create(fpath, use_first=False)

    # This should work now
    node4, create = usp.UspData.get_or_create(fpath, use_first=True)
    assert create is False
    assert node4.pk in (node1.pk, node2.pk)


def test_usp_element_validation(new_database, usp_folder):
    """Test the validation mechanism"""

    import aiida_castep.data.usp as usp
    fpath = str(usp_folder / "Sr_00.usp")
    # Pass a inconsistent element should not work
    with pytest.raises(ValidationError):
        usp.UspData.get_or_create(fpath, element="Ti", store_usp=True)

    fpath = usp_folder / "foo.usp"
    fpath.write_text(six.text_type("adfalal"))
    # This should work since the element is defined explicitly
    usp.UspData.get_or_create(fpath, element="Ti", store_usp=True)

    # This should also work
    fpath = usp_folder / "bar.usp"
    fpath.write_text(six.text_type("asdfddf"))
    pp, _ = usp.UspData.get_or_create(fpath, store_usp=False)
    pp.set_element("Ti")
    pp.store()



"""
Calculations of CASTEP
"""
from __future__ import print_function
from __future__ import absolute_import
import six
from six.moves import zip

from aiida.common import InputValidationError
from aiida.common import CalcInfo, CodeInfo
from aiida.plugins import DataFactory
from aiida.engine import ProcessBuilder
from aiida.orm.nodes.data.base import to_aiida_type

from aiida.orm import UpfData, User
from aiida.engine import CalcJob

from ..common import INPUT_LINKNAMES, OUTPUT_LINKNAMES, EXIT_CODES_SPEC
from .inpgen import CastepInputGenerator
from ..data.otfg import OTFGData
from ..data.usp import UspData
from .utils import get_castep_ion_line
from .tools import (castep_input_summary, update_parameters,
                    use_pseudos_from_family, input_param_validator,
                    check_restart)

from .._version import CALC_PARSER_VERSION
__version__ = CALC_PARSER_VERSION

KpointsData = DataFactory("array.kpoints")
StructureData = DataFactory("structure")
Dict = DataFactory("dict")

inp_ln = INPUT_LINKNAMES
out_ln = OUTPUT_LINKNAMES
ecodes = EXIT_CODES_SPEC

# Define the version of the calculation

__all__ = ['CastepCalculation', 'submit_test']


class CastepCalculation(CalcJob, CastepInputGenerator):
    """
    Class representing a generic CASTEP calculation -
    This class should work for all types of calculations.
    """

    # Create a dict of the defaults
    _DEFAULTS = {
        "seedname": 'aiida',
        'symlink_usage': True,
        'parent_folder_name': 'parent',
        'parser_name': 'castep.castep',
        'use_kpoints': True,
        'withmpi': True,
    }
    _DEFAULTS['input_filename'] = _DEFAULTS['seedname'] + '.cell'
    _DEFAULTS['output_filename'] = _DEFAULTS['seedname'] + '.castep'

    _default_retrieve_list = [
        "*.err", "*.den_fmt", "*.elf_fmt", "*-out.cell", "*.pdos_bin"
    ]

    # Some class methods
    retrieve_dict = {
        "phonon": [".phonon"],
        "phonon+efield": [".phonon", ".efield"],
        "magres": [".magres"],
        "transitionstatesearch": [".ts"],
        "molecular dynamics": [".md"],
        "moleculardynamics": [".md"],
        "geometryoptimisation": [".geom"],
        "geometryoptimization": [".geom"],
        "spectral": [".ome_bin", ".dome_bin"],
    }

    # NOT CURRENTLY USED
    _acceptable_tasks = [
        "singlepoint",
        "geometryoptimization",
        "geometryoptimisation",
    ]

    _copied_attributes = [
        "jobresource_param", "custom_scheduler_commands",
        "max_wallclock_seconds"
    ]
    _write_headers = True

    _cell_links = [
        inp_ln['parameters'], inp_ln['structure'], inp_ln['settings'],
        inp_ln['kpoints']
    ]

    _param_links = [inp_ln['parameters']]

    # Extra kpoints - CASTEP has many calculation mode that take extra kpoints
    _extra_kpoints = {
        'spectral': {  # name XX_kpoints_list
            'task': ('spectra', ),
            'need_weights':
            True  # Whether the explicit kpoints need weights or not
        },
        'bs': {
            'task': ('bandstructure', ),  # task where the kpoints will be used
            'need_weigthts': False,
        },
        'phonon': {
            'task': ('phonon', 'phonon+efield'),
            'need_weights': False,
        },
        'phonon_fine': {
            'task': ('phonon', 'phonon+efield'),
            'need_weights': False,
        },
        'supercell': {
            'task': ('phonon', ),
            'need_weights': True,
        },
        'magres': {
            'task': ('magres', ),
            'need_weights': True,
        },
        'optics': {
            'task': ('optics', ),
            'need_weights': True,
        },
        'elnes': {
            'task': ('elnes', ),
            'need_weights': True,
        }
    }

    @classmethod
    def define(cls, spec):
        import aiida.orm as orm
        super(CastepCalculation, cls).define(spec)

        # Initialise interal params, saved as metadata.options
        for key, value in cls._DEFAULTS.items():
            port_name = 'metadata.options.' + key
            spec.input(port_name, default=value)

        spec.input('metadata.options.retrieve_list',
                   valid_type=list,
                   default=cls._default_retrieve_list)

        # Begin defining the input nodes
        spec.input(inp_ln['structure'],
                   valid_type=orm.StructureData,
                   help="Defines the input structure")
        spec.input(inp_ln['settings'],
                   valid_type=orm.Dict,
                   serializer=to_aiida_type,
                   required=False,
                   help="Use an additional node for sepcial settings")
        spec.input(inp_ln['parameters'],
                   valid_type=orm.Dict,
                   serializer=to_aiida_type,
                   validator=input_param_validator,
                   help="Use a node that sepcifies the input parameters")
        # TODO: implement logic to automaticall set the restart if such folder is given
        spec.input(
            inp_ln['parent_calc_folder'],
            valid_type=orm.RemoteData,
            help=
            'Use a remote folder as the parent folder. Useful for restarts.',
            required=False)
        spec.input_namespace(
            'pseudos',
            valid_type=(UspData, OTFGData, UpfData),
            help=("Use nodes for the pseudopotentails of one of"
                  "the element in the structure. You should pass a"
                  "a dictionary specifying the pseudpotential node for"
                  "each kind such as {O: <PsudoNode>}"),
            dynamic=True)
        spec.input(inp_ln['kpoints'],
                   valid_type=KpointsData,
                   required=False,
                   help="Use a node defining the kpoints for the calculation")

        # Define additional kpoints for different tasks
        for key, value in cls._extra_kpoints.items():
            tasks = ', '.join(value['task'])
            spec.input(key + '_' + inp_ln['kpoints'],
                       valid_type=KpointsData,
                       required=False,
                       help="Extra kpoints input for task: {}".format(tasks))

        # Define the exit codes
        for smsg, (code, msg, inv) in ecodes.items():
            spec.exit_code(code, smsg, message=msg, invalidates_cache=inv)

        # Define the output nodes
        spec.output(out_ln['results'],
                    required=True,
                    valid_type=Dict,
                    help='Parsed results in a dictionary format.')

        spec.outputs.dynamic = True
        # Define the default inputs, enable CalcJobNode to use .res
        spec.default_output_node = out_ln['results']

    def prepare_for_submission(self, folder):
        """
        Routine to be called when create the input files and other stuff

        :param folder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputs_dict (without the Code!)
        """
        self.prepare_inputs()

        local_copy_list = []
        remote_copy_list = []
        remote_symlink_list = []

        # TODO allow checking of inputs
        #if self.inputs.metadata.options._auto_input_validation is True:
        #    self.check_castep_input(self.param_dict, auto_fix=False)

        # If requested to reuse, check if the parent_calc_folder is defined
        require_parent = False
        for k in self.param_dict:
            if str(k).lower() in ["reuse", "continuation"]:
                require_parent = True
                break

        parent_calc_folder = self.inputs.get('parent_calc_folder')
        if parent_calc_folder is None and require_parent:
            raise InputValidationError(
                "No parent calculation folder passed"
                " for restart calculation using reuse/continuation")

        ##############################
        # END OF INITIAL INPUT CHECK #
        ##############################

        # Generate input file
        self.prepare_inputs(reset=True)

        if self._write_headers is True:
            cell_nodes = []
            for name, inp in self.inputs.items():
                if name in self._cell_links and inp:
                    cell_nodes.append([name, inp])

            # process pseudos
            for name, pseudo in self.inputs.pseudos.items():
                cell_nodes.append(['pseudo__{}'.format(name), pseudo])

            self.cell_file.header = self._generate_header_lines(cell_nodes)

            param_nodes = []
            for name, inp in self.inputs.items():
                if name in self._param_links and inp:
                    param_nodes.append([name, inp])

            self.param_file.header = self._generate_header_lines(param_nodes)

        local_copy_list.extend(self.local_copy_list_to_append)
        seedname = self.inputs.metadata.options.seedname

        cell_fn = seedname + ".cell"
        param_fn = seedname + ".param"

        with folder.open(cell_fn, mode='w') as incell:
            incell.write(six.text_type(self.cell_file.get_string()))

        with folder.open(param_fn, mode="w") as inparam:
            inparam.write(six.text_type(self.param_file.get_string()))

        # IMPLEMENT OPERATIONS FOR RESTART

        symlink = self.inputs.metadata.options.symlink_usage
        parent_calc_folder = self.inputs.get('parent_calc_folder', None)
        if parent_calc_folder:
            comp_uuid = parent_calc_folder.computer.uuid
            remote_path = parent_calc_folder.get_remote_path()
            if symlink:
                remote_list = remote_symlink_list
            else:
                remote_list = remote_copy_list
            remote_list.append(
                (comp_uuid, remote_path,
                 self.inputs.metadata.options.parent_folder_name))

        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid

        # COPY/SYMLINK LISTS
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.remote_symlink_list = remote_symlink_list

        # SET UP extra CMDLINE arguments
        cmdline_params = self.settings_dict.pop("CMDLINE", [])

        # Extra parameters are added after the seed for CASTEP
        calcinfo.cmdline_params = [seedname] + list(cmdline_params)

        # CASTEP don't have any STDOUT etc when running calculations
        # Error is shown in the *.err file

        # Construct codeinfo instance
        codeinfo = CodeInfo()
        codeinfo.cmdline_params = [seedname] + list(cmdline_params)
        codeinfo.code_uuid = self.inputs.code.uuid
        calcinfo.codes_info = [codeinfo]

        # Retrieve by default the .castep file and the bands file
        calcinfo.retrieve_list = []
        calcinfo.retrieve_list.append(seedname + ".castep")
        calcinfo.retrieve_list.append(seedname + ".bands")

        settings_retrieve_list = self.settings_dict.pop(
            "ADDITIONAL_RETRIEVE_LIST", [])
        calcinfo.retrieve_list.extend(settings_retrieve_list)

        calculation_mode = self.param_file.get("task", "singlepoint")

        # If we are doing geometryoptimisation retrieved the geom file and -out.cell file
        # dictionary for task specific file retrieve
        task_extra = self.retrieve_dict.get(calculation_mode.lower(), [])
        for suffix in task_extra:
            settings_retrieve_list.append(seedname + suffix)

        # Retrieve output cell  file if requested
        if self.param_file.get("write_cell_structure"):
            settings_retrieve_list.append(seedname + "-out.cell")

        calcinfo.retrieve_list += settings_retrieve_list
        calcinfo.retrieve_list += self._default_retrieve_list

        # Remove parser options in the setting dictionary
        # At the moment parser options are not used here

        if self.settings_dict:
            raise InputValidationError(
                "The following keys have been found in "
                "the settings input node, but were not understood: {}".format(
                    ",".join(list(self.settings_dict.keys()))))

        return calcinfo

    # Attach the input summary method
    @staticmethod
    def get_castep_input_summary(builder):
        """Summarize the input for a builder"""
        return castep_input_summary(builder)

    @classmethod
    def submit_test(cls, *args, **kwargs):
        """Test submission with a builder of inputs"""
        if args and isinstance(args[0], ProcessBuilder):
            return submit_test(args[0])
        else:
            return submit_test(cls, **kwargs)

    @classmethod
    def check_restart(cls, builder, verbose=False):
        """Check the existence of restart file is needed"""
        check_restart(builder, verbose)

    @classmethod
    def dryrun_test(cls, inputs, castep_exe='castep.serial', verbose=True):
        """
        Do a dryrun test in a folder with prepared builder or inputs
        """

        from fnmatch import fnmatch
        from aiida.common.folders import Folder
        if isinstance(inputs, ProcessBuilder):
            res = cls.submit_test(inputs)
        else:
            res = cls.submit_test(cls, **inputs)
        folder = Folder(res[1])
        dry_run_node = res[0]
        seedname = dry_run_node.get_option('seedname')

        def _print(inp):
            if verbose:
                print(inp)

        # Do a dryrun
        from subprocess import call, check_output
        try:
            output = check_output([castep_exe, "-v"], universal_newlines=True)
        except OSError:
            _print("CASTEP executable '{}' is not found".format(castep_exe))
            return

        # Now start dryrun
        _print("Running with {}".format(
            check_output(["which", castep_exe], universal_newlines=True)))
        _print(output)

        _print("Starting dryrun...")
        call([castep_exe, "--dryrun", seedname], cwd=folder.abspath)

        # Check if any *err files
        contents = folder.get_content_list()
        for n in contents:
            if fnmatch(n, "*.err"):
                with folder.open(n) as fh:
                    _print("Error found in {}:\n".format(n))
                    _print(fh.read())
                raise InputValidationError("Error found during dryrun")

        # Gather information from the dryrun file
        import re
        dryrun_results = {}
        out_file = seedname + '.castep'
        with folder.open(out_file) as fh:
            for line in fh:
                mth = re.match(r"\s*k-Points For SCF Sampling:\s+(\d+)\s*",
                               line)
                if mth:
                    dryrun_results["num_kpoints"] = int(mth.group(1))
                    _print("Number of k-points: {}".format(mth.group(1)))
                    mth = None
                    continue
                mth = re.match(
                    r"\| Approx\. total storage required"
                    r" per process\s+([0-9.]+)\sMB\s+([0-9.]+)", line)
                if mth:
                    dryrun_results["memory_MB"] = (float(mth.group(1)))
                    dryrun_results["disk_MB"] = (float(mth.group(2)))
                    _print("RAM: {} MB, DISK: {} MB".format(
                        mth.group(1), mth.group(2)))
                    mth = None
                    continue

        return folder, dryrun_results

    def _prepare_cell_file(self):
        """Add extra kpoints information to the calculation"""
        # First, call the base method
        super(CastepCalculation, self)._prepare_cell_file()
        param = self.inputs.get(inp_ln['parameters']).get_dict()
        task = param['PARAM'].get('task', 'singlepoint')

        # Check if we have more kpoints
        for kpn_name, kpn_settings in self._extra_kpoints.items():
            extra_kpns = self.inputs.get(kpn_name + '_' + inp_ln['kpoints'])
            # No need to proceed if it is not defined
            if extra_kpns is None:
                continue
            self._include_extra_kpoints(extra_kpns, kpn_name, kpn_settings)
            # Warn if this kpoint will not be used by the task
            if task not in kpn_settings['task']:
                self.report(
                    'Warning: kpoints for {} will not be used for task {}'.
                    format(kpn_name, task))

    @staticmethod
    def update_paraemters(inputs, *args, **kwargs):
        """Update the paramters for a given input dictionary/builder"""
        return update_parameters(inputs, *args, **kwargs)

    @staticmethod
    def use_pseudos_from_family(inputs, family_name):
        use_pseudos_from_family(inputs, family_name)

    def _generate_header_lines(self, other_nodes=None):
        """
        Generate header lines to go into param and cell files
        :param other_nodes: A list of pairs of (linkname, node)

        """
        from textwrap import TextWrapper
        import time
        from aiida.manage.manager import get_manager
        profile = get_manager().get_profile()
        if not profile:
            return

        wrapper = TextWrapper(initial_indent="# ", subsequent_indent="# ")
        time_str = time.strftime("%H:%M:%S %d/%m/%Y %Z")
        lines = [
            "##### Generated by aiida_castep {} #####".format(time_str),
            "#         author: Bonan Zhu (bz240@cam.ac.uk)",
            "# "
            "# AiiDA User: {}".format(
                User.objects.get_default().get_full_name()),
            "# AiiDA profile: {}".format(profile.name),
            "# Information of the calculation node",
            #"# type: {}".format(self.get_name()),
            #"# pk: {}".format(self.pk),
            #"# uuid: {}".format(self.uuid),
            "# label: {}".format(self.inputs.metadata.get('label')),
            "# description:",
        ]

        description = self.inputs.metadata.get('description')
        if description:
            lines.extend(wrapper.wrap(description))
        lines.append("")

        # additional information of the input nodes
        if other_nodes:
            lines.append("## Information of input nodes used:")

        for name, node in other_nodes:
            node_lines = [
                "# ", "# type: {}".format(node), "# pk: {}".format(node.pk),
                "# linkname: {}".format(name), "# uuid: {}".format(node.uuid),
                "# label: {}".format(node.label), "# description:"
            ]
            _desc = node.description
            if _desc:
                node_lines.extend(wrapper.wrap(_desc))
            node_lines.append("")
            lines.extend(node_lines)

        lines.append("# END OF HEADER")

        return lines


class TaskSpecificCalculation(CastepCalculation):
    """
    Class for Calculations that only allow certain tasks
    """

    _acceptable_tasks = []

    def prepare_for_submission(self, folder):

        in_dict = self.inputs[INPUT_LINKNAMES['parameters']].get_dict()

        # Check if task is correctly set
        all_tasks = [t.lower() for t in self._acceptable_tasks]
        if in_dict['PARAM']['task'].lower() not in all_tasks:
            raise InputValidationError("Wrong TASK value {}"
                                       " set in PARAM".format(
                                           in_dict['PARAM']['task'].lower()))
        return super(TaskSpecificCalculation,
                     self).prepare_for_submission(folder)


class CastepTSCalculation(TaskSpecificCalculation):
    """
    CASTEP calculation for transition state search. Use an extra input product structure.
    """
    _acceptable_tasks = ["transitionstatesearch"]

    @classmethod
    def define(cls, spec):
        import aiida.orm as orm
        super(CastepTSCalculation, cls).define(spec)
        spec.input(inp_ln['prod_structure'],
                   valid_type=orm.StructureData,
                   required=True,
                   help='Product structure for transition state search.')

    def _prepare_cell_file(self):
        """
        Extend the prepare_cell_filer method to include product
        structure
        """
        super(CastepTSCalculation, self)._prepare_cell_file()
        p_structure = self.inputs[inp_ln['prod_structure']]
        pdt_position_list = []
        for site in p_structure.sites:
            kind = p_structure.get_kind(site.kind_name)
            name = kind.symbol
            line = get_castep_ion_line(name, site.position)
            pdt_position_list.append(line)

        self.cell_file["POSITIONS_ABS_PRODUCT"] = pdt_position_list


def submit_test(arg, **kwargs):
    """This essentially test the submition"""
    from aiida.engine import run_get_node

    # Deal with passing an process builder
    if isinstance(arg, ProcessBuilder):
        inputs = arg

        inputs['metadata']['store_provenance'] = False
        inputs['metadata']['dry_run'] = True

        output_node = run_get_node(inputs).node
        inputs['metadata']['store_provenance'] = True
        inputs['metadata']['dry_run'] = False
    else:
        inputs = kwargs
        inputs['metadata']['store_provenance'] = False
        inputs['metadata']['dry_run'] = True
        output_node = run_get_node(arg, **inputs).node
        inputs['metadata']['store_provenance'] = True
        inputs['metadata']['dry_run'] = False

    return output_node, output_node.dry_run_info['folder']



"""
Test for generating castep input
"""
from __future__ import absolute_import
from __future__ import print_function
import os
import unittest

from aiida.common import InputValidationError
from aiida.common.folders import SandboxFolder
from aiida.plugins import DataFactory, CalculationFactory
from aiida.backends.testbase import AiidaTestCase
from aiida.orm import Code
from aiida_castep.calculations.castep import CastepTSCalculation as TSCalc
from .dbcommon import BaseDataCase, BaseCalcCase

CasCalc = CalculationFactory("castep.castep")
BSCalc = CalculationFactory("castep.castep")
TSCalc = CalculationFactory("castep.ts")
StructureData = DataFactory("structure")
Dict = DataFactory("dict")
KpointsData = DataFactory("array.kpoints")


class TestCastepInputGeneration(AiidaTestCase, BaseCalcCase, BaseDataCase):
    """
    Test if the input is correctly generated
    """
    @classmethod
    def setUpClass(cls):
        super(TestCastepInputGeneration, cls).setUpClass()
        cls.calc_params = {
            "computer": cls.computer,
            "resources": {
                "num_machines": 1,
                "num_mpiprocs_per_machine": 1
            }
        }

        cls.code = Code()
        cls.code.set_remote_computer_exec((cls.computer, "/x.x"))
        cls.code.store()

    def make_H2_structure(self):

        cell = ((5., 0., 0.), (0., 5., 0.), (0., 0., 5.))
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["H"])
        s.append_atom(position=(1., 0., 0.), symbols=["H"])
        self.H2 = s

    def test_pre_submit_checkings(self):
        """
        Test checkup before submission
        """
        pass

    def test_using_OTFG_family(self):
        """
        Test using OTFG in the input
        """
        from .utils import get_STO_structure
        STO = get_STO_structure()
        full, missing, C9 = self.create_family()
        c = CasCalc()
        pdict = self.get_default_input()
        # pdict["CELL"].pop("block species_pot")
        p = Dict(dict=pdict).store()
        c.use_structure(STO)
        c.use_pseudos_from_family(full)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_code(self.code)

        input_dict = c.get_inputs_dict()
        # Check mixing libray with acutal entry
        self.assertEqual(input_dict["pseudo_O"].entry, "C9")

        with SandboxFolder() as f:
            p = Dict(dict=pdict)
            c.use_parameters(p)
            input_dict = c.get_inputs_dict()
            c._prepare_for_submission(f, input_dict)

    def test_using_OTFG_mix(self):
        """
        Test using mixed OTFG library and manual values
        """
        pass

    def test_using_UpfData(self):
        """
        Test using UpfData
        """
    def generate_test_calc(self):
        """
        Return a defined Calculation
        """

        cell = ((2., 0., 0.), (0., 2., 0.), (0., 0., 2.))

        input_params = {
            "PARAM": {
                "task": "singlepoint",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
            }
        }

        c = CasCalc(**self.calc_params).store()
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["Ba"])
        s.append_atom(position=(1., 0., 0.), symbols=["Ba"])
        s.store()

        p = Dict(dict=input_params).store()

        k = KpointsData()
        k.set_kpoints_mesh([4, 4, 4])
        k.store()

        settings_dict = {"SPINS": [0, 0]}
        c.use_settings(Dict(dict=settings_dict))
        c.label = "TEST CALC"
        c.description = "Test calculation for AiiDA CASTEP plugin."
        c.use_code(self.code)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_structure(s)
        c.use_parameters(p)

        return c

    def test_inputs(self):

        cell = ((2., 0., 0.), (0., 2., 0.), (0., 0., 2.))

        input_params = {
            "PARAM": {
                "task": "singlepoint",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                "species_pot": ("Ba Ba_00.usp", )
            }
        }

        c = CasCalc(**self.calc_params).store()
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["Ba"])
        s.append_atom(position=(1., 0., 0.), symbols=["Ba"])
        s.store()

        p = Dict(dict=input_params).store()

        k = KpointsData()
        k.set_kpoints_mesh([4, 4, 4], offset=(0.5, 0.5, 0.5))

        settings_dict = {"SPINS": [0, 0]}
        c.use_settings(Dict(dict=settings_dict))
        c.label = "TEST CALC"
        c.description = ("Test calculation for AiiDA CASTEP plugin. "
                         "Testing generation of inputs.")
        inputdict = c.get_inputs_dict()

        inputdict.pop("code", None)

        with SandboxFolder() as f:
            # I use the same SandboxFolder more than once because nothing
            # should be written for these failing tests

            # Missing required input nodes
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_parameters(p)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_structure(s)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_kpoints(k)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)

            # Continuation keywords present - should raise exception
            inputdict = c.get_inputs_dict()
            input_params["PARAM"]["continuation"] = "default"
            p2 = Dict(dict=input_params)
            inputdict["parameters"] = p2
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)

            c.use_code(self.code)
            inputdict = c.get_inputs_dict()
            c._prepare_for_submission(f, inputdict)

            # Check existenc of the file
            cell = f.get_abs_path(c._SEED_NAME + ".cell", check_existence=True)
            param = f.get_abs_path(c._SEED_NAME + ".param",
                                   check_existence=True)
            print("\n" + "#" * 5 + "CONTENT OF CELL FILE: " + "#" * 5)

            # Check and print out the input files
            with open(cell) as p:
                cell_content = p.read()
                print(cell_content)
                self.assertIn("fix_all_cell", cell_content)
                self.assertIn("kpoints_mp_offset", cell_content)
                self.assertIn("kpoints_mp_grid", cell_content)

            print("\n" + "#" * 5 + "CONTENT OF PARAM FILE: " + "#" * 5)
            with open(param) as p:
                param_content = p.read()
                print(param_content)
                self.assertIn("task", param_content)

            k.set_kpoints_mesh((2, 2, 2), (0, 0, 0))
            c.use_kpoints(k)
            inputdict = c.get_inputs_dict()
            c._prepare_for_submission(f, inputdict)
            with open(cell) as p:
                cell_content = p.read()
                print(cell_content)
                self.assertNotIn("kpoints_mp_offset", cell_content)

    @unittest.skip("Dry run test takes a long time.")
    def test_dryrun(self):
        from subprocess import call

        try:
            call(["castep.serial", "-v"])
        except OSError:
            self.skipTest("No CASTEP excutable found")

        c = self.generate_test_calc()
        full, missing, C9 = self.create_family()
        c.use_pseudos_from_family(C9)
        # Do a dry run - check if any error message is given
        with SandboxFolder() as f:
            c._prepare_for_submission(f, c.get_inputs_dict())
            self.castep_dryrun(f, c._SEED_NAME)
            self.assertFalse(f.get_content_list("*.err"))
            self.assertTrue(f.get_content_list("*.castep"))

    def castep_dryrun(self, folder, seed):
        from subprocess import call
        import os
        seed = os.path.join(folder.abspath, seed)
        call(["castep.serial", seed, "-dryrun"], cwd=folder.abspath)


class TestRestartGeneration(AiidaTestCase, BaseCalcCase, BaseDataCase):
    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestRestartGeneration, cls).setUpClass(*args, **kwargs)
        cls.setup_code_castep()

    def test_restart(self):

        # Initial logic of creation
        c1 = self.setup_calculation()
        c1.store_all()
        with self.assertRaises(InputValidationError):
            c2 = c1.create_restart(ignore_state=False)
        with self.assertRaises(InputValidationError):
            c2 = c1.create_restart(ignore_state=True, reuse=True)

        rmd = self.get_remote_data("H2-geom")
        rmd.store()
        c1._set_state("RETRIEVING")
        from aiida.common.links import LinkType
        rmd.add_link_from(c1, link_type=LinkType.CREATE)

        # This simply create a copy of c1
        c2 = c1.create_restart(ignore_state=True)
        c2_inp = c2.get_inputs_dict()
        c1_inp = c1.get_inputs_dict()

        # Check all inputs of c1 are indeed present in c2
        for key in c1_inp:
            self.assertIn(key, c2_inp)
            self.assertEqual(c2_inp[key].pk, c1_inp[key].pk)

        c2 = c1.create_restart(ignore_state=True, reuse=True)
        c2_inp = c2.get_inputs_dict()
        c1_inp = c1.get_inputs_dict()

        c2_param = c2_inp[c2.get_linkname("parameters")].get_dict()['PARAM']

        reuse = c2_param.pop("reuse")
        self.assertEqual(
            reuse,
            os.path.join(c2._restart_copy_to,
                         "{}.check".format(c2._SEED_NAME)))

        with SandboxFolder() as f:
            print(c2._prepare_for_submission(f, c2_inp))

    def test_continue_from(self):
        """
        Test the continue_from function.
        It calls the same underlying function compared with
        create_restart
        """
        c1 = self.setup_calculation()
        c1.store_all()
        c2 = BSCalc.continue_from(c1, ignore_state=True, reuse=False)

        rmd = self.get_remote_data("H2-geom")
        rmd.store()
        c1._set_state("RETRIEVING")
        from aiida.common.links import LinkType
        rmd.add_link_from(c1, link_type=LinkType.CREATE)

        c3 = BSCalc.continue_from(c1, ignore_state=True)
        c1_inp = c1.get_inputs_dict()
        c2_inp = c2.get_inputs_dict()
        c3_inp = c3.get_inputs_dict()

        # Check all inputs of c1 are indeed present in c2
        for key in c1_inp:
            self.assertIn(key, c2_inp)

        for key in c1_inp:
            self.assertIn(key, c3_inp)

        with SandboxFolder() as f:
            c3._prepare_for_submission(f, c3_inp)


class TestTSCalculation(BaseCalcCase, BaseDataCase, AiidaTestCase):
    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestTSCalculation, cls).setUpClass(*args, **kwargs)
        cls.setup_code_castep()

    def get_default_input(self):

        input_params = {
            "PARAM": {
                "task": "transitionstatesearch",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                # "species_pot": ("Ba Ba_00.usp",)
            }
        }

        return input_params

    def setup_calculation(self, param=None):
        from .utils import get_STO_structure

        STO = get_STO_structure()
        full, missing, C9 = self.create_family()
        c = TSCalc()

        if param:
            pdict = param
        else:
            pdict = self.get_default_input()

        # pdict["CELL"].pop("block species_pot")
        p = Dict(dict=pdict)
        c.use_structure(STO)
        c.use_product_structure(STO)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_code(self.code)
        c.set_computer(self.computer)
        c.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 2})
        c.use_parameters(p)

        # Check mixing libray with acutal entry
        return c

    def test_input_validation(self):
        """Test input validation"""

        c = self.setup_calculation()
        pdict = c.get_inputs_dict()[c.get_linkname('parameters')].get_dict()
        pdict['PARAM']['task'] = "singlepoint"
        c.get_inputs_dict()[c.get_linkname('parameters')].set_dict(pdict)

        with SandboxFolder() as f:
            with self.assertRaises(InputValidationError):
                inputs = c.get_inputs_dict()
                c._prepare_for_submission(f, inputs)

    def test_generation(self):
        """
        Test if the block has been generated
        """
        c = self.setup_calculation()
        with SandboxFolder() as f:
            inputs = c.get_inputs_dict()
            c._prepare_for_submission(f, inputs)
            with f.open("aiida.cell") as fcell:
                cell = fcell.read()
                self.assertIn("%BLOCK POSITIONS_ABS_PRODUCT", cell)
                self.assertIn("%ENDBLOCK POSITIONS_ABS_PRODUCT", cell)


class TestBSCalculation(BaseCalcCase, BaseDataCase, AiidaTestCase):
    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestBSCalculation, cls).setUpClass(*args, **kwargs)
        cls.setup_code_castep()

    def get_default_input(self):

        input_params = {
            "PARAM": {
                "task": "bandstructure",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                # "species_pot": ("Ba Ba_00.usp",)
            }
        }

        return input_params

    def get_bs_kpoints(self):
        kpoints = KpointsData()
        kpoints.set_kpoints([[0, 0, 0], [0.5, 0.5, 0.5]])
        return kpoints

    def setup_calculation(self, param=None):
        from .utils import get_STO_structure

        STO = get_STO_structure()
        full, missing, C9 = self.create_family()
        c = BSCalc()

        if param:
            pdict = param
        else:
            pdict = self.get_default_input()

        # pdict["CELL"].pop("block species_pot")
        p = Dict(dict=pdict)
        c.use_structure(STO)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_bs_kpoints(self.get_bs_kpoints())
        c.use_code(self.code)
        c.set_computer(self.computer)
        c.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 2})
        c.use_parameters(p)

        # Check mixing libray with acutal entry
        return c

    def test_input_validation(self):
        """Test input validation"""

        c = self.setup_calculation()
        pdict = c.get_inputs_dict()[c.get_linkname('parameters')].get_dict()
        pdict['PARAM']['task'] = "singlepoint"
        c.get_inputs_dict()[c.get_linkname('parameters')].set_dict(pdict)

        with SandboxFolder() as f:
            with self.assertRaises(InputValidationError):
                inputs = c.get_inputs_dict()
                c._prepare_for_submission(f, inputs)

        c = self.setup_calculation()
        c.CHECK_EXTRA_KPN = True

        with SandboxFolder() as f:
            with self.assertRaises(InputValidationError):
                inputs = c.get_inputs_dict()
                inputs.pop("bs_kpoints")
                c._prepare_for_submission(f, inputs)

    def test_restart(self):
        """
        Test creating an restart of the calculation.
        The restart should retain the extra kpoints inputs.
        """
        c = self.setup_calculation()
        new = c.create_restart(ignore_state=True)
        self.assertEqual(type(new), type(c))
        inpname = "{}_kpoints".format(c.kpn_name)
        self.assertIn(new.get_linkname(inpname), new.get_inputs_dict())

        # Extra kpn is not mandatory
        extra_kpn = c.get_inputs_dict()["bs_kpoints"]
        c._remove_link_from("bs_kpoints")
        self.assertNotIn(extra_kpn, c.get_inputs())
        # This should still succeed
        new = c.create_restart(ignore_state=True)
        self.assertNotIn(extra_kpn, new.get_inputs())

    def test_bs_kpoints(self):

        c = self.setup_calculation()
        inputs = c.get_inputs_dict()
        with SandboxFolder() as f:
            c._prepare_for_submission(f, inputs)
            with f.open("aiida.cell") as cell:
                content = cell.read()
        self.assertIn("%BLOCK BS_KPOINTS_LIST", content)

    def test_bs_kpoints_mp(self):

        c = self.setup_calculation()
        inputs = c.get_inputs_dict()
        mp = KpointsData()
        mp.set_kpoints_mesh([2, 2, 2], (0.1, 0.1, 0.1))
        inputs['bs_kpoints'] = mp

        with SandboxFolder() as f:
            c._prepare_for_submission(f, inputs)
            with f.open("aiida.cell") as cell:
                content = cell.read()
        self.assertIn("{:<20}: 2 2 2".format("bs_kpoints_mp_grid"), content)
        self.assertIn("{:<20}: 0.1 0.1 0.1".format("bs_kpoints_mp_offset"),
                      content)



"""
Test input generation
"""
from __future__ import absolute_import
import pytest
from aiida_castep.calculations.inpgen import CastepInputGenerator
from aiida.engine.processes.ports import PortNamespace


@pytest.fixture
def gen_instance():
    """
    Create a mock generator intstance
    """
    gen = CastepInputGenerator()
    gen.inputs = PortNamespace
    gen.inputs.metadate = PortNamespace
    gen.inputs.metadate.options = PortNamespace
    return gen


@pytest.fixture
def param_dict():
    """Example dict in Dict"""

    out = {
        "PARAM": {
            "task": "singlepoints",
            "xc_functional": "lda"
        },
        "CELL": {
            "fix_all_cell": True,
            "cell_constraints": ["1 1 1", "2 2 3"]
        }
    }
    return out


@pytest.mark.parametrize('entry_point', ('castep.castep', 'castep.ts'))
def test_get_builder(db_test_app, entry_point):
    from aiida.plugins import CalculationFactory
    cls = CalculationFactory(entry_point)
    builder = cls.get_builder()


def test_inp_gen_param(gen_instance, param_dict):
    """
    Test generate paramters
    """
    gen_instance.param_dict = param_dict
    gen_instance._prepare_param_file()
    assert gen_instance.param_file == param_dict["PARAM"]


def test_inp_gen_cell(gen_instance, sto_calc_inputs):
    """
    Test generation of the inputs
    """
    gen_instance.inputs = sto_calc_inputs
    gen_instance.prepare_inputs()
    assert 'symmetry_generate' in gen_instance.cell_file
    assert "POSITIONS_ABS" in gen_instance.cell_file
    assert "LATTICE_CART" in gen_instance.cell_file
    assert isinstance(gen_instance.cell_file["cell_constraints"], list)
    assert 'C9' in gen_instance.cell_file['SPECIES_POT'][0]

    # Test extra-kpoints
    from aiida.orm import KpointsData
    kpn1 = KpointsData()
    kpn1.set_kpoints_mesh((4, 4, 4))
    gen_instance._include_extra_kpoints(kpn1, 'phonon', {
        'task': ('phonon', ),
        'need_weights': False
    })
    assert 'phonon_kpoint_mp_grid' in gen_instance.cell_file

    kpn1.set_kpoints_mesh((
        4,
        4,
        4,
    ), (0.25, 0.25, 0.25))
    gen_instance._include_extra_kpoints(kpn1, 'phonon', {
        'task': ('phonon', ),
        'need_weights': False
    })
    assert 'phonon_kpoint_mp_offset' in gen_instance.cell_file

    kpn2 = KpointsData()
    kpn_points = [[0, 0, 0], [0.5, 0.5, 0.5]]
    kpn_weights = [0.3, 0.6]
    kpn2.set_kpoints(kpn_points, weights=kpn_weights)
    gen_instance._include_extra_kpoints(kpn2, 'bs', {
        'task': ('bandstructure', ),
        'need_weights': True
    })
    assert 'BS_KPOINT_LIST' in gen_instance.cell_file


def test_cell_with_tags(gen_instance, sto_calc_inputs):
    """
    Test that the inputs generator correctly handles the case with
    kind.name != symbol
    """
    from ..utils import get_mixture_cell
    sto_tags = get_mixture_cell()
    sto_calc_inputs.structure = sto_tags
    pseudos = sto_calc_inputs.pseudos
    pseudos['O1'] = pseudos['O']
    pseudos['O2'] = pseudos['O']
    pseudos['SrTi_Sr'] = pseudos['O']
    pseudos['SrTi_Ti'] = pseudos['O']
    gen_instance.inputs = sto_calc_inputs
    gen_instance.prepare_inputs()

    positions = gen_instance.cell_file['POSITIONS_ABS']

    assert 'Sr:SrTi' in positions[0].split('\n')[0]
    assert 'Ti:SrTi' in positions[0].split('\n')[1]

    # Check the currect type are there
    species_pots = [
        tuple(line.split()) for line in gen_instance.cell_file['SPECIES_POT']
    ]
    assert ('Sr:SrTi', 'C9') in species_pots
    assert ('O:O1', 'C9') in species_pots
    assert ('O:O2', 'C9') in species_pots


@pytest.mark.process_execution
def test_submission(new_database, sto_calc_inputs, sto_spectral_inputs):
    """
    Test submitting a CastepCalculation
    """
    from aiida_castep.calculations.castep import CastepCalculation
    from aiida.engine import run_get_node
    _, return_node = run_get_node(CastepCalculation, **sto_calc_inputs)
    assert return_node.exit_status == 106  # No castep output found

    # test with extra kpoints
    _, return_node = run_get_node(CastepCalculation, **sto_spectral_inputs)
    assert return_node.exit_status == 106


def test_submit_test(new_database, sto_calc_inputs):
    """
    Test the ``submit_test`` method
    """
    from aiida_castep.calculations.castep import CastepCalculation
    from aiida.common.folders import Folder
    builder = CastepCalculation.get_builder()
    builder._update(sto_calc_inputs)
    res = CastepCalculation.submit_test(builder)
    fcontent = Folder(res[1]).get_content_list()
    assert 'aiida.cell' in fcontent
    assert 'aiida.param' in fcontent


def test_submit_test_function(new_database, sto_calc_inputs):
    """
    Test the ``submit_test`` method
    """
    from aiida_castep.calculations.castep import CastepCalculation, submit_test
    from aiida.common.folders import Folder

    # Test with process class and inputs
    res = submit_test(CastepCalculation, **sto_calc_inputs)
    fcontent = Folder(res[1]).get_content_list()
    assert 'aiida.cell' in fcontent
    assert 'aiida.param' in fcontent
    # Nothing should change for the nested dic
    assert sto_calc_inputs['metadata'].get('dry_run') is not True
    assert sto_calc_inputs['metadata'].get('store_provenance') is not False

    # Test with builder
    builder = CastepCalculation.get_builder()
    builder._data = sto_calc_inputs
    res = submit_test(builder)
    fcontent = Folder(res[1]).get_content_list()
    assert 'aiida.cell' in fcontent
    assert 'aiida.param' in fcontent

    # Nothing should change in the process builder
    assert builder.metadata.get('dry_run') is not True
    assert builder.metadata.get('store_provenance') is not False


def test_param_validation(db_test_app):
    """Test input validations"""
    from aiida_castep.calculations.castep import CastepCalculation

    builder = CastepCalculation.get_builder()
    # This shold work
    builder.parameters = {'PARAM': {'cut_off_energy': 100}}

    with pytest.raises(ValueError):
        # Flat style is not allowed
        builder.parameters = {'foo': 'bar'}

    with pytest.raises(ValueError):
        # Erorr will be catached
        builder.parameters = {'PARAM': {'cut_off_eneryg': 100}}


def run_castep_calc(inputs):
    from aiida_castep.calculations.castep import CastepCalculation
    from aiida.engine import run_get_node
    return run_get_node(CastepCalculation, **inputs)[1]


def test_dict2builder(aiida_profile, sto_calc_inputs):
    """Test that we can use nested dict input for builder"""
    from aiida_castep.calculations.castep import CastepCalculation
    from aiida.engine import run_get_node
    builder = CastepCalculation.get_builder()
    builder._update(sto_calc_inputs)
    run_get_node(builder)



"""
Tests for calculation module
"""
from __future__ import absolute_import
import pytest
from aiida_castep.common import INPUT_LINKNAMES, OUTPUT_LINKNAMES


@pytest.fixture
def calcjobnode(h2_calc_inputs, generate_calc_job_node):
    """Create a fake calcjob node"""
    inp_structure = h2_calc_inputs[INPUT_LINKNAMES['structure']]
    calcjobnode = generate_calc_job_node(
        'castep.castep',
        'H2-geom',
        inputs=h2_calc_inputs,
        outputs={OUTPUT_LINKNAMES['structure']: inp_structure.clone()})

    return calcjobnode


def test_use_pseudos(sto_calc_inps_or_builder, create_otfg_group):
    """Test the `use_psudos_from_family` function"""
    from aiida_castep.calculations.tools import use_pseudos_from_family
    create_otfg_group(['QC5'], 'QC5')
    use_pseudos_from_family(sto_calc_inps_or_builder, 'QC5')
    assert all([
        pseudo.entry == 'QC5'
        for pseudo in sto_calc_inps_or_builder['pseudos'].values()
    ])


def test_castep_summary_builder(sto_calc_inps_or_builder):
    """Test the get_castep_input_summary method"""
    from aiida_castep.calculations.tools import castep_input_summary

    keys = [
        "kpoints", "structure", "code", "computer", "resources",
        "custom_scheduler_commands", "wallclock", "label", "pseudos"
    ]
    out_dict = castep_input_summary(sto_calc_inps_or_builder)
    for k in keys:
        assert k in out_dict


def test_castep_summary_calcjob(sto_calc_inputs, generate_calc_job_node):
    """Test the summary method works for CalcJobNode"""
    from aiida_castep.calculations.tools import castep_input_summary
    from tempfile import mkdtemp
    calcjobnode = generate_calc_job_node(entry_point_name='castep.castep',
                                         results_folder=mkdtemp(),
                                         inputs=sto_calc_inputs)
    out_dict = castep_input_summary(calcjobnode)

    keys = [
        "kpoints", "structure", "code", "computer", "resources",
        "custom_scheduler_commands", "wallclock", "label", "pseudos"
    ]
    for k in keys:
        assert k in out_dict


def test_param_update(sto_calc_inps_or_builder):
    """Test the param update function, it should work for both
    inputs dictionary and builders"""
    from aiida_castep.calculations.tools import update_parameters
    from aiida_castep.common import INPUT_LINKNAMES

    inputs = sto_calc_inps_or_builder
    out = update_parameters(inputs, xc_functional='scan')

    assert out is inputs
    assert inputs[INPUT_LINKNAMES['parameters']].get_dict(
    )['PARAM']['xc_functional'] == 'scan'

    # Test deletion
    out = update_parameters(inputs, delete=['xc_functional'])
    assert 'xc_functional' not in inputs[
        INPUT_LINKNAMES['parameters']].get_dict()['PARAM']

    # Test assestion of stored node
    inputs[INPUT_LINKNAMES['parameters']].store()
    with pytest.raises(RuntimeError):
        out = update_parameters(inputs, xc_functional='scan')


def test_create_restart_builder(sto_calc_inputs):
    """Test creating restart from a builder"""
    from aiida_castep.calculations.tools import create_restart
    new_builder = create_restart(sto_calc_inputs, entry_point='castep.castep')

    new_builder = create_restart(sto_calc_inputs,
                                 entry_point='castep.castep',
                                 param_update={'fix_all_cell': True})
    assert new_builder.parameters.get_dict()['CELL']['fix_all_cell'] is True

    # Test deletion
    new_builder = create_restart(new_builder, param_delete=['fix_all_cell'])
    assert 'fix_all_cell' not in new_builder.parameters.get_dict()['CELL']


def test_create_restart_node(calcjobnode):
    """Test creating a restart from a CalcJobNode"""

    with pytest.raises(RuntimeError):
        new_builder = calcjobnode.tools.create_restart(
            False, param_update={'fix_all_cell': True})

    new_builder = calcjobnode.tools.create_restart(
        True, param_update={'fix_all_cell': True})
    assert new_builder.parameters.get_dict()['CELL']['fix_all_cell'] is True

    new_builder = calcjobnode.tools.create_restart(True,
                                                   use_output_structure=True)
    assert new_builder[
        INPUT_LINKNAMES['structure']].uuid == calcjobnode.outputs.__getattr__(
            OUTPUT_LINKNAMES['structure']).uuid



"""
Commandline plugin module for aiida CASTEP plugin
"""
import click

from aiida.cmdline.commands.cmd_data import verdi_data

# pylint: disable=import-outside-toplevel


@verdi_data.group('castep-pseudos')
def pseudos_cmd():
    """Commandline interface for working with OTFGData"""
@pseudos_cmd.command(name="migrate-families")
def migrate_families():
    """
    Migrate to new style Group for pseudopotential families as implemented in
    `aiida-core >= 1.2.0`.

    This command should be run once after upgrading `aiida-castep` to 1.2.0,
    to re-enable working with pseudopotential families.
    """
    from aiida_castep.data.otfg import migrate_otfg_family
    click.echo("Migrating to new style Group")
    migrate_otfg_family()
    click.echo(
        "Finished! Note: No need to run this command again for this profile.")


@pseudos_cmd.command(name="listfamilies")
@click.option(
    '--element',
    '-e',
    multiple=True,
    help=
    "Show families contenting this element only. Can be passed multiple times")
@click.option('--with_description', '-d', is_flag=True)
def listfamilies(element, with_description):
    """List available OtfgData families"""
    from aiida.orm import QueryBuilder, Node
    from aiida_castep.data.otfg import OTFGGroup

    qbd = QueryBuilder()
    qbd.append(Node, tag="otfgdata")
    if element:
        qbd.add_filter("otfgdata", {
            "attributes.element": {
                "or": [{
                    'in': element
                }, {
                    '==': "LIBRARY"
                }]
            }
        })
    qbd.append(OTFGGroup,
               tag='group',
               with_node='otfgdata',
               project=['label', 'description'])
    qbd.distinct()
    if qbd.count() > 0:
        for res in qbd.dict():
            group_label = res.get("group").get("label")
            group_desc = res.get("group").get("description")
            # Count the number of pseudos in this group
            qbd = QueryBuilder()
            qbd.append(OTFGGroup,
                       tag='thisgroup',
                       filters={"label": {
                           'like': group_label
                       }})
            qbd.append(Node, project=["id"], with_group='thisgroup')

            if with_description:
                description_string = ": {}".format(group_desc)
            else:
                description_string = ""

            click.echo("* {} [{} pseudos]{}".format(group_label, qbd.count(),
                                                    description_string))

    else:
        click.echo("No valid pseudopotential family found.")



from __future__ import absolute_import
from aiida_castep.calculations.helper import CastepHelper, HelperCheckError
import unittest

helper = CastepHelper()
no_info = helper.BY_PASS


class TestHelper(unittest.TestCase):
    """class  TestHelper for test CastepHepler"""
    @classmethod
    def setUpClass(cls):
        cls.helper = CastepHelper()

    @property
    def flat_dict(self):
        return dict(fix_all_cell="true",
                    cut_off_energy="true",
                    kpoints_mp_grid="0 0 0")

    @property
    def only_param_dict(self):
        return dict(PARAM={"cut_off_energy": "100", "nextra_bands": "200"})

    @property
    def only_cell_dict(self):
        return dict(CELL={"fix_all_cell": "100", "symmetry_generate": "200"})

    @property
    def input_dict(self):
        d = self.only_cell_dict
        d.update(self.only_param_dict)
        return d

    @unittest.skipIf(no_info, "No helper info found")
    def test_from_flat(self):
        out, not_found = self.helper._from_flat_dict(self.flat_dict)
        # print(self.helper.help_dict)
        self.assertFalse(not_found)

    @unittest.skipIf(no_info, "No helper info found")
    def test_check_dict_raw(self):
        """Test the underlying check_dict function"""

        invalid, wrong = self.helper._check_dict(self.only_cell_dict)
        self.assertFalse(any([wrong, invalid]))
        invalid, wrong = self.helper._check_dict(self.only_param_dict)
        self.assertFalse(any([wrong, invalid]))

        comb = self.only_cell_dict
        comb.update(self.only_param_dict)
        invalid, wrong = self.helper._check_dict(comb)
        self.assertFalse(any([wrong, invalid]))

        comb["CELL"].update(kpointx_mp_grid="d")
        comb["PARAM"].update(snap_to_symmetry="true")
        invalid, wrong = self.helper._check_dict(comb)
        self.assertIn("kpointx_mp_grid", invalid)
        self.assertIn(("snap_to_symmetry", "CELL"), wrong)

    @unittest.skipIf(no_info, "No helper info found")
    def test_check_dict(self):
        """Test top level check dict function"""

        # Pass a good dictionary
        comb = self.input_dict
        dout = self.helper.check_dict(comb)
        self.assertEqual(dout, comb)

        # Pass a wrong dictioanry
        comb.update(FOO="bla")
        with self.assertRaises(HelperCheckError):
            self.helper.check_dict(comb)

        # Pass a mix and matched
        comb.pop("FOO")
        comb.update({"spin": "1"})
        outdict = self.helper.check_dict(comb)
        self.assertIn("spin", outdict["PARAM"])

        with self.assertRaises(HelperCheckError):
            outdict = self.helper.check_dict(comb, auto_fix=False)

        comb["CELL"].update({"spin_polarised": "true"})
        outdict = self.helper.check_dict(comb, auto_fix=True)
        self.assertIn("spin_polarised", outdict["PARAM"])

        # Check incompatible keys can be spotted
        outdict["PARAM"].update(elec_method="dm")
        self.helper.check_dict(outdict, auto_fix=False)

        outdict["PARAM"].pop("elec_method")
        outdict["PARAM"].update(metals_method="dm")
        self.helper.check_dict(outdict, auto_fix=False)

        # Should raise error if both keys present
        outdict["PARAM"].update(elec_method="dm")
        with self.assertRaises(HelperCheckError):
            outdict = self.helper.check_dict(outdict, auto_fix=False)

    @unittest.skipIf(no_info, "No helper info found")
    def test_check_dict_flat(self):
        """Test checking error in a flat dictionary"""
        flat = self.flat_dict
        self.helper.check_dict(flat, auto_fix=False, allow_flat=True)
        with self.assertRaises(HelperCheckError):
            flat['kpoints_gdrd'] = 1
            self.helper.check_dict(flat, auto_fix=False, allow_flat=True)


if __name__ == "__main__":
    unittest.main()



"""
Tests for utils module
"""

from __future__ import absolute_import
import pytest
from ..utils import *
import numpy as np

try:
    import ase
except ImportError:
    ase = None


@pytest.fixture
def unsorted_atoms():
    atoms = ase.Atoms("TiO2",
                      cell=[5, 5, 5],
                      positions=[[0, 0, 0], [1, 0, 0], [0, 1, 0]])
    return atoms


@pytest.fixture
def sorted_atoms():
    atoms = ase.Atoms(numbers=[8, 8, 22],
                      cell=[5, 5, 5],
                      positions=[[1, 0, 0], [0, 1, 0], [0, 0, 0]])
    return atoms


@pytest.mark.skipif(ase is None, reason="No ase module")
def test_ase_to_castep_index(unsorted_atoms):
    res = ase_to_castep_index(unsorted_atoms, [0, 2, 1])
    assert res[0] == ["Ti", 1]
    assert res[1] == ["O", 2]
    assert res[2] == ["O", 1]


@pytest.mark.skipif(ase is None, reason="No ase module")
def test_sort_atoms(unsorted_atoms, sorted_atoms):
    unsorted_atoms = sort_atoms_castep(unsorted_atoms, order=None)
    assert np.all(unsorted_atoms.numbers == sorted_atoms.numbers)
    assert np.all(unsorted_atoms.positions == sorted_atoms.positions)


@pytest.mark.skipif(ase is None, reason="No ase module")
def test_desort_atoms(unsorted_atoms, sorted_atoms):
    sorted_atoms = sort_atoms_castep(unsorted_atoms, order=None)
    tmp = desort_atoms_castep(sorted_atoms, unsorted_atoms)
    assert tmp == unsorted_atoms


@pytest.mark.skipif(ase is None, reason="No ase module")
def test_check_sorted(unsorted_atoms, sorted_atoms):
    assert is_castep_sorted(unsorted_atoms) == False
    assert is_castep_sorted(sorted_atoms) == True


def test_k_spacing():
    spacing = compute_kpoints_spacing([1, 1, 1], [1, 1, 1])
    assert np.all(spacing == np.array([1, 1, 1]))

    spacing = compute_kpoints_spacing([4, 4, 4], [2, 2, 2])
    assert np.all(spacing == np.array([1. / 8, 1. / 8, 1. / 8]))



"""
Tests for datastructure
"""
from __future__ import absolute_import
import pytest
from aiida_castep.calculations.datastructure import CastepInputFile


@pytest.fixture
def basic_input():
    c = CastepInputFile()
    c["a"] = "a"
    c["b"] = ["a", "b"]
    c["c"] = 5
    return c


def test_input_gen(basic_input):
    """
    Test basic function of generate inputs
    """
    lines = basic_input.get_file_lines()
    assert lines[0].split(":")[0].strip() == "a"
    assert lines[0].split(":")[1].strip() == "a"
    assert lines[1].startswith("%")
    assert lines[4].startswith("%")


def test_header(basic_input):
    """
    Test adding header
    """
    basic_input.header = ["Hello World"]
    lines = basic_input.get_file_lines()
    assert lines[0] == "# Hello World"

    basic_input.header = ["#Hello World"]
    lines = basic_input.get_file_lines()
    assert lines[0] == "#Hello World"


def test_unit(basic_input):
    """
    Test the unit system
    """
    basic_input.units["a"] = "eV"
    basic_input.units["b"] = "eV"
    lines = basic_input.get_file_lines()
    assert lines[0].split(":")[1].strip()[-2:] == "eV"
    assert lines[2].strip() == "eV"


def test_string(basic_input):
    lines = basic_input.get_file_lines()
    "\n".join(lines) == basic_input.get_string()



"""
This module stores a single version for calculation-parser.

A single version number is used here so whenever any change to parser
 or calculation modules, the version number should
be increased.This avoids falsely matching nodes via hashes.

CHANGELOG
0.2.3 -> FIX a typo psedu_pots -> pseudo_pots
0.2.4 -> Sort the bands and parsed kpoints from .bands file using the index given

1.0.0 -> initial version for plugin version 1.0.
Changed the warning messages to EXIT_CODE style

1.0.1 -> Revision to include error messages
Added parsing of the *.err file and put the content in the field 'error_messages'
of the returned Dict.
The order of error is also adjusted. Fixed a bug where if the *.err files are
present the returned code would be same as when end of the calculation is not
found. This fix allows the internal crash of CASTEP (usually with the *.err files)
to be differentiated with the calculation being killed (by scheduler).

1.0.2 -> Fix a bug of parsing cons'd forces

1.0.3 -> Fix a bug where the forces are not reorderred as for the case of StructureData
"""

CALC_PARSER_VERSION = "1.0.3"
PLUGIN_VERSION = "1.2.0a2"



"""
Common module
Courtesy to aiida-quantumespresso's WorkChain design
"""
from __future__ import absolute_import
from aiida.common import AiidaException
from collections import namedtuple
from functools import wraps
from aiida.engine import ExitCode


class UnexpectedCalculationFailure(AiidaException):
    """Exception for encountering unknown errors"""
    pass


ErrorHandler = namedtuple('ErrorHandler', 'priority method')
"""
A namedtuple to define an error handler for a :class:`~aiida.engine.processes.workchains.workchain.WorkChain`.

The priority determines in which order the error handling methods are executed, with
the higher priority being executed first. The method defines an unbound WorkChain method
that takes an instance of a :class:`~aiida.orm.nodes.process.calculations.calcjob.CalcJobNode`
as its sole argument. If the condition of the error handler is met, it should return an :class:`.ErrorHandlerReport`.

:param priority: integer denoting the error handlers priority
:param method: the workchain class method
"""

ErrorHandlerReport = namedtuple('ErrorHandlerReport',
                                'is_handled do_break exit_code')
ErrorHandlerReport.__new__.__defaults__ = (False, False, ExitCode())
"""
A namedtuple to define an error handler report for a :class:`~aiida.engine.processes.workchains.workchain.WorkChain`.

This namedtuple should be returned by an error handling method of a workchain instance if
the condition of the error handling was met by the failure mode of the calculation.
If the error was appriopriately handled, the 'is_handled' field should be set to `True`,
and `False` otherwise. If no further error handling should be performed after this method
the 'do_break' field should be set to `True`

:param is_handled: boolean, set to `True` when an error was handled, default is `False`
:param do_break: boolean, set to `True` if no further error handling should be performed, default is `False`
:param exit_code: an instance of the :class:`~aiida.engine.processes.exit_code.ExitCode` tuple
"""


def register_error_handler(cls, priority):
    """
    Decorator that will turn any function in an error handler for workchain that inherits from
    the :class:`.BaseCastepWorkChain`. The function expects two arguments, a workchain class and a priortity.
    The decorator will add the function as a class method to the workchain class and add an :class:`.ErrorHandler`
    tuple to the :attr:`.BaseCastepWorkChain._error_handlers` attribute of the workchain. During failed calculation
    handling the :meth:`.inspect_calculation` outline method will call the `_handle_calculation_failure` which will loop
    over all error handler in the :attr:`.BaseCastepWorkChain._error_handlers`, sorted with respect to the priority in
    reverse. If the workchain class defines a :attr:`.BaseCastepWorkChain._verbose` attribute and is set to `True`, a
    report message will be fired when the error handler is executed.

    Requirements on the function signature of error handling functions. The function to which the
    decorator is applied needs to take two arguments:

        * `self`: This is the instance of the workchain itself
        * `calculation`: This is the calculation that failed and needs to be investigated

    The function body should usually consist of a single conditional that checks the calculation if
    the error that it is designed to handle is applicable. Although not required, it is advised that
    the function return an :class:`.ErrorHandlerReport` tuple when its conditional was met. If an error was handled
    it should set `is_handled` to `True`. If no other error handlers should be considered set `do_break` to `True`.

    :param cls: the workchain class to register the error handler with
    :param priority: an integer that defines the order in which registered handlers will be called
        during the handling of a failed calculation. Higher priorities will be handled first
    """
    def error_handler_decorator(handler):
        @wraps(handler)
        def error_handler(self, calculation):
            if hasattr(cls, '_verbose') and cls._verbose:
                self.report('({}){}'.format(priority, handler.__name__))
            return handler(self, calculation)

        setattr(cls, handler.__name__, error_handler)

        if not hasattr(cls, '_error_handlers'):
            cls._error_handlers = []
        cls._error_handlers.append(ErrorHandler(priority, error_handler))

        return error_handler

    return error_handler_decorator



"""
CASTEP HELPER
Check for errors in input dictionary
{"CELL": {...},
 "PARAM": {...}

"""

from __future__ import absolute_import
from __future__ import print_function
import os
import json
from glob import glob

import logging
from six.moves import zip
logger = logging.getLogger(__name__)

path = os.path.abspath(__file__)
module_path = os.path.dirname(path)

incompatible_keys = [
    ('backup_interval', 'num_backup_iter'),
    ('opt_strategy', 'opt_strategy_bias'),
    ('basis_precision', 'cut_off_energy'),
    ('fine_grid_scale', 'fine_gmax', 'find_cut_off_energy'),
    ('nelectrons', 'charge'),
    ('nelectrons', 'nup'),
    ('nelectrons', 'ndown'),
    ('charge', 'ndown'),
    ('charge', 'nup'),
    ('spin', 'nup'),
    ('spin', 'ndown'),
    ('nextra_bands', 'perc_extra_bands'),
    ('metals_method', 'elec_method'),
]


class HelperCheckError(RuntimeError):
    pass


class CastepHelper(object):
    """
    A class for helping castep inputs
    """
    def __init__(self, version=None):
        # Instance level parameter for by passing any check at all
        self.BY_PASS = False
        self._HELP_DICT = None
        self.version = version
        # Try to load th helper dictionary
        self.load_helper_dict()

    def load_helper_dict(self):
        """
        Attempts to load a dictionary containing the helper information.
        This dictionary should be generated by the .generate module
        We look for .castep_helpinfo.json in the $HOME directory.
        """
        pairs = find_help_info()
        path = None
        for p, v in pairs:
            if v == self.version:
                path = p
                break
        # If we did not found a suitable version
        if path is None:
            if self.version is not None:
                raise RuntimeError(
                    "Cannot found help info for requested version {}".format(
                        self.version))
            try:
                path = pairs[0][
                    0]  # No explicitly requested - use the first one
            except IndexError:
                print("No CASTEP help info detected")
                self.BY_PASS = True
                help_dict = {}
            else:
                help_dict = load_json(path)
        else:
            help_dict = load_json(path)
        # Save the help_dict as a class attribute
        self._HELP_DICT = help_dict

    @property
    def help_dict(self):
        """A dictionary containing information of the helper"""
        return self._HELP_DICT

    @property
    def castep_help_version(self):
        """Version number of CASTEP that the help info is for"""
        return self.help_dict["_CASTEP_VERSION"]

    def save_helper_dict(self, help_dict, file_path):
        """
        Save the helper json file in to sensible location. By default it is here this module is
        """

        try:
            with open(file_path, "w") as fp:
                json.dump(help_dict, fp)
        except OSError:
            try:
                with ("castep_helpinfo.json", "w") as fp:
                    json.dump(help_dict, fp)
                    file_path = os.path.realpath("castep_helpinfo.json")
                    print(("Saving in current path. "
                           "Please move it to {}".format(file_path)))
            except OSError:
                print("Cannot save the retrieved help information")
                return

        print(("\n\nJSON file saved in {}".format(file_path)))

    def _check_dict(self, input_dict):
        """
        Check a dictionary of inputs. Return invalid and wrong keys

        :param dict input_dict: A dictonary with "PARAM" and "CELL" keys.

        :return: A list of [invalid_keys, wrong_keys].
        wrong_keys is a tuple of where the key should have being.
        (key, "PARAM") or (key, "CELL")
        """
        invalid_keys = []
        wrong_keys = []  # a list of tulple (key, "PARAM") or (key, "CELL")

        if self.BY_PASS:
            return invalid_keys, wrong_keys

        for kwtype in input_dict:
            # Check each key
            for key in input_dict[kwtype]:
                # key maybe be both lower or upper case
                info = self.help_dict.get(key.lower(), None)

                if info is None:
                    invalid_keys.append(key)
                    continue

                # Check if the type is correct
                if info["key_type"] != kwtype:
                    wrong_keys.append((key, info["key_type"]))
                    continue

        return invalid_keys, wrong_keys

    def _from_flat_dict(self, input_dict):
        """
        Construct a {"PARAM":{}, "CELL":{}} dictionary from dictionary with top-level keys other
        than PARAM and CELL
        :returns: a list of [out_dict, keys_not_found]
        """

        if self.BY_PASS:
            raise RuntimeError(
                "Cannot construct dictionary - No help info found")

        hinfo = self.help_dict

        # extract copies of CELL and PARAM fields from the input
        cell_dict = dict(input_dict.get("CELL", {}))
        param_dict = dict(input_dict.get("PARAM", {}))
        not_found = []
        for key in input_dict:
            key_entry = hinfo.get(key, None)

            if key_entry is None:
                not_found.append(key)
            else:
                kwtype = key_entry.get("key_type")
                if kwtype == "CELL":
                    cell_dict.update({key: input_dict[key]})
                elif kwtype == "PARAM":
                    param_dict.update({key: input_dict[key]})
                else:
                    raise RuntimeError(
                        "Entry {} does not have key_type value".format(key))

        out_dict = {"CELL": cell_dict, "PARAM": param_dict}
        return out_dict, not_found

    def check_dict(self, input_dict, auto_fix=True, allow_flat=False):
        """
        Check input dictionary. Apply and warn about errors
        :param input_dict dictionary: a dictionary as the input, contain "CELL", "PARAM" and other keywords
        :param auto_fix bool: Whether we should fix error automatically
        :param allow_flat: Accept that the input dictionary is flat.

        :returns dict: A structured dictionary
        """
        input_dict = input_dict.copy()  # this is a shallow copy

        # construct what to be checked
        cell_dict = input_dict.pop("CELL", {})
        param_dict = input_dict.pop("PARAM", {})

        if input_dict and not auto_fix and not allow_flat:
            raise HelperCheckError("keywords: {} at top level".format(
                ", ".join(input_dict)))

        # Following functions require help info to be defined
        if self.BY_PASS:
            logger.warning("No help info found - input not checked")
            return input_dict

        # process what's left
        re_structured, not_found = self._from_flat_dict(input_dict)
        if not_found:
            suggests = [self.get_suggestion(s) for s in not_found]
            # Warnings
            not_reco = [
                "keyword '{}' is not recognized at top level".format(s)
                for s in not_found
            ]
            # Suggestions
            sugst_str = [a + "\n" + b for a, b in zip(not_reco, suggests)]
            # Combine warnings and suggestions together
            sugst_str = "\n\n".join(sugst_str)
            raise HelperCheckError(sugst_str)

        # Now construct a dictionary
        cell_dict.update(re_structured["CELL"])
        param_dict.update(re_structured["PARAM"])

        # Construct an new dictionary
        input_dict = dict(CELL=cell_dict, PARAM=param_dict)

        # Check the restructed dictionary
        invalid, wrong = self._check_dict(input_dict)

        if invalid:
            suggests = [self.get_suggestion(s) for s in invalid]
            not_founds = [
                "keyword '{}' is not found".format(s) for s in invalid
            ]
            sugst_str = [a + "\n" + b for a, b in zip(not_founds, suggests)]
            sugst_str = "\n\n".join(sugst_str)

            raise HelperCheckError(sugst_str)

        # if there are still wrong keywords, fix them
        if wrong:
            if auto_fix is True:
                for key, should_be in wrong:
                    if should_be == "PARAM":
                        logger.warning("Key {} moved to PARAM".format(key))
                        value = input_dict["CELL"].pop(key)
                        input_dict["PARAM"].update({key: value})
                    else:
                        logger.warning("Key {} moved to CELL".format(key))
                        value = input_dict["PARAM"].pop(key)
                        input_dict["CELL"].update({key: value})
            else:
                raise HelperCheckError(
                    "Keywords: {} are in "
                    "the wrong sub-dictionary".format(", ".join(
                        ["'{}'".format(k[0]) for k in wrong])))

        # Check incompatible keys
        incomp = check_incompatible(input_dict["PARAM"], incompatible_keys)
        if incomp:
            raise HelperCheckError("Incompatible keys found: {}"
                                   " - only one is allowed.".format(incomp))
        return input_dict

    def get_suggestion(self, string):
        """
        Return string for suggestion of the string
        """
        return _get_suggestion(string, list(self.help_dict.keys()))


def _get_suggestion(provided_string, allowed_strings):
    """
    Given a string and a list of allowed_strings, it returns a string to print
    on screen, with sensible text depending on whether no suggestion is found,
    or one or more than one suggestions are found.

    :param provided_string: the string to compare

    :param allowed_strings: a list of valid strings

    :return: A string to print on output, to suggest to the user
    a possible valid value.
    """
    import difflib

    similar_kws = difflib.get_close_matches(provided_string, allowed_strings)
    if len(similar_kws) == 1:
        return "(Maybe you wanted to specify {0}?)".format(similar_kws[0])
    elif len(similar_kws) > 1:
        return "(Maybe you wanted to specify one of these: {0}?)".format(
            ", ".join(similar_kws))
    else:
        return "(No similar keywords found...)"


def load_json(path):
    """
    Load a json via file path
    """
    with open(path) as f:
        res = json.load(f)
    return res


def check_incompatible(dict_in, inc_list):
    """
    Check any conflicting keys in the dictionary
    """
    for incomp in inc_list:
        c = 0
        for k in dict_in:
            if k in incomp:
                c += 1
        if c > 1:
            return incomp
    return


def find_help_info():
    """
    Return possible paths of helper dict
    The diction should store in the format:
    castep_help_info_<version>.json
    Search paths are the module path and the $HOME

    :returns: [path1, path2], [verion1, version2].....
    """
    home = os.getenv("HOME")
    default_file_path = os.path.join(module_path, "*castep_help_info_*.json")
    others = glob(os.path.join(home, ".*castep_help_info_*.json"))
    paths = glob(default_file_path) + others
    vs = []
    for p in paths:
        tmp = p.replace(".json", "").split("_")
        if len(tmp) == 4:
            vs.append(float(tmp[-1]))
        else:
            vs.append(0)
    comb = list(zip(paths, vs))
    comb.sort(key=lambda x: x[1], reverse=True)
    return comb



"""
In this test we adopt the PluginTestCase provided by aiida and migrate
existing AiiDATestCase tests with minimum effort
"""

from __future__ import absolute_import
import io
import os

from aiida.common.folders import SandboxFolder
from aiida.common import ValidationError
from unittest import TestCase
import pytest

Ti_otfg = "Ti 3|1.8|9|10|11|30U:40:31:32(qc=5.5)"
Sr_otfg = "Sr 3|2.0|5|6|7|40U:50:41:42"
O_otfg = "O 2|1.1|15|18|20|20:21(qc=7)"

# This fixture will be run for each class and ensure
# AiiDA test environment has been loaded correctly


@pytest.fixture(scope="class")
def import_things(aiida_profile, request):
    import aiida_castep.data.otfg as otf
    import aiida_castep.data.usp as usp
    from aiida.plugins import DataFactory
    request.cls.otfg = otf.OTFGData
    request.cls.otf = otf
    request.cls.usp = usp
    request.cls.usp = usp


@pytest.mark.usefixtures("import_things")
class BaseDataCase(TestCase):
    """Base to include some useful things"""

    # @pytest.fixture(autouse=True, scope="function")
    # def populate_env(self):
    #     """
    #     Popoulate the class namespace with the imported modules
    #     """
    #     if not self.AIIDA_ENV_LOADED:
    #         import aiida_castep.data.otfg as otf
    #         import aiida_castep.data.usp as usp
    #         from aiida.orm import DataFactory
    #         self.otfg = DataFactory("castep.otfgdata")
    #         self.otf = otf
    #         self.usp = usp
    #         self.AIIDA_ENV_LOADED = True

    @pytest.fixture(autouse=True)
    def reset_db(self, aiida_profile):
        aiida_profile.reset_db()
        yield
        aiida_profile.reset_db()

    def create_family(self):
        """Creat families for testsing"""
        Ti, Sr, O, C9 = self.get_otfgs()
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, O.entry], "STO_FULL",
                                    "TEST", False)
        self.otf.upload_otfg_family([Ti.entry, Sr.entry], "STO_O_missing",
                                    "TEST", False)

        # Missing O but that's OK we have a C9 wild card here
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, "C9"], "STO_O_C9",
                                    "TEST", False)

        return "STO_FULL", "STO_O_missing", "STO_O_C9"

    def get_otfgs(self):

        Ti, _ = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        Sr, _ = self.otfg.get_or_create(Sr_otfg, store_otfg=False)
        O, _ = self.otfg.get_or_create(O_otfg, store_otfg=False)
        C9, _ = self.otfg.get_or_create("C9", store_otfg=False)
        return Ti, Sr, O, C9

    @staticmethod
    def get_STO_structure():
        """Return a STO structure"""
        from aiida.plugins import DataFactory
        StructureData = DataFactory("structure")
        a = 3.905

        cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["Sr"])
        s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
        s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
        s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"])
        s.append_atom(position=(0., a / 2, a / 2), symbols=["O"])
        s.label = "STO"
        return s


class TestOTFGData(BaseDataCase):
    def setUp(self):
        self.otfg_nodes = {}

    def test_otfg_split(self):
        element, setting = self.otf.split_otfg_entry(Ti_otfg)
        self.assertEqual(element, "Ti")
        self.assertEqual(setting, "3|1.8|9|10|11|30U:40:31:32(qc=5.5)")

    def test_otfg_create(self):

        element, setting = self.otf.split_otfg_entry(Ti_otfg)
        C9 = self.otfg(otfg_entry="C9")
        self.assertEqual(C9.string, "C9")
        self.assertEqual(C9.element, "LIBRARY")
        C9.store()

        Ti = self.otfg(otfg_entry=Ti_otfg)
        self.assertEqual(Ti.string, setting)
        self.assertEqual(Ti.element, element)
        Ti.store()

    def test_get_or_create(self):

        Ti, create = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        self.assertTrue(create)

        # Create but not stored do it again should have no change
        Ti, create = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        self.assertTrue(create)

        Ti.store()

        Ti2, create = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        self.assertFalse(create)

        # Should get the stored OTFG entry
        # If using SQLA this Ti should be IS Ti2?
        self.assertEqual(Ti2.uuid, Ti.uuid)

        # Using another way should be the same
        Sr = self.otfg()
        Sr.set_element("Sr")
        Sr.set_string("bla")
        Sr.store()

        Sr2, create = self.otfg.get_or_create("Sr_bla")
        self.assertFalse(create)

        # A different OTFG
        Sr3, create = self.otfg.get_or_create("Sr_foo")
        self.assertTrue(create)

        # Check if more than one instance is found in the db
        with self.assertRaises(ValueError):
            Sr = self.otfg()
            Sr.set_element("Sr")
            Sr.set_string("bla")
            Sr.store()

            # This should fail there are more than one
            Sr4, create = self.otfg.get_or_create("Sr_bla", use_first=False)

        Sr4, create = self.otfg.get_or_create("Sr_bla", use_first=True)

    def test_set_up_family(self):

        Ti, Sr, O, C9 = self.get_otfgs()

        otfgs = [Ti.entry, Sr.entry, C9.entry]
        entry, uploaded = self.otf.upload_otfg_family(otfgs, "Test", "Test")
        self.assertEqual((entry, uploaded), (3, 3))

        # This should fail
        with self.assertRaises(ValidationError):
            entry, uploaded = self.otf.upload_otfg_family(
                otfgs, "Test", "Test")

        entry, uploaded = self.otf.upload_otfg_family([O.entry] + otfgs,
                                                      "Test",
                                                      "Test",
                                                      stop_if_existing=False)

        groups = self.otfg.get_otfg_groups()
        self.assertEqual(len(groups), 1)

        retrieved_entries = [node.entry for node in groups[0].nodes]
        for o in otfgs + [O.entry]:
            self.assertIn(o, retrieved_entries)

    def test_assign_from_structure(self):
        """
        Test using get_pseudos_from_structure
        """

        from aiida_castep.data import get_pseudos_from_structure
        from aiida.common import NotExistent

        self.create_family()
        STO = self.get_STO_structure()

        pseudo_list = get_pseudos_from_structure(STO, "STO_FULL")
        self.assertEqual(pseudo_list["Sr"].entry, Sr_otfg)
        self.assertEqual(pseudo_list["O"].entry, O_otfg)
        self.assertEqual(pseudo_list["Ti"].entry, Ti_otfg)

        with self.assertRaises(NotExistent):
            pseudo_list = get_pseudos_from_structure(STO, "STO_O_missing")

        pseudo_list = get_pseudos_from_structure(STO, "STO_O_C9")
        self.assertEqual(pseudo_list["Sr"].entry, Sr_otfg)
        self.assertEqual(pseudo_list["O"].entry, "C9")
        self.assertEqual(pseudo_list["Ti"].entry, Ti_otfg)

    def create_family(self):
        """Creat families for testsing"""
        Ti, Sr, O, C9 = self.get_otfgs()
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, O.entry], "STO_FULL",
                                    "TEST", False)
        self.otf.upload_otfg_family([Ti.entry, Sr.entry], "STO_O_missing",
                                    "TEST", False)

        # Missing O but that's OK we have a C9 wild card here
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, "C9"], "STO_O_C9",
                                    "TEST", False)

    def get_otfgs(self):

        Ti, _ = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        Sr, _ = self.otfg.get_or_create(Sr_otfg, store_otfg=False)
        O, _ = self.otfg.get_or_create(O_otfg, store_otfg=False)
        C9, _ = self.otfg.get_or_create("C9", store_otfg=False)
        return Ti, Sr, O, C9


class TestUspData(BaseDataCase):
    def upload_usp_family(self):
        """Make a fake usp node"""

        with SandboxFolder() as f:
            sub = f.get_subfolder("pseudo", create=True)
            for element in ["Sr", "Ti", "O"]:
                fp = io.StringIO(u"foo bla 42")
                sub.create_file_from_filelike(fp,
                                              "{}_00.usp".format(element),
                                              mode='w')

            self.usp.upload_usp_family(os.path.join(f.abspath, "pseudo"),
                                       "STO", "")

            with self.assertRaises(ValueError):
                self.usp.upload_usp_family(os.path.join(f.abspath, "pseudo"),
                                           "STO", "")

    def get_usp_node(self, element):
        """
        Return a node of usp file
        """
        name = "{}_00.usp".format(element)
        with SandboxFolder() as f:
            fp = io.StringIO(u"foo bla 42")
            f.create_file_from_filelike(fp, name, mode='w')
            fpath = os.path.join(f.abspath, name)
            node = self.usp.UspData.get_or_create(fpath)[0]

        return node

    def test_get_or_create(self):
        """Testing the logic or get_or_create"""
        name = "Sr_00.usp"
        with SandboxFolder() as f:
            fp = io.StringIO(u"foo bla 42")
            f.create_file_from_filelike(fp, name, mode='w')
            fpath = os.path.join(f.abspath, name)
            node1, create = self.usp.UspData.get_or_create(fpath)

            self.assertTrue(create)
            self.assertEqual(node1.element, "Sr")

            node2 = self.usp.UspData(file=fpath)
            node2.store()

            # Now having two files - should raise an exception
            with self.assertRaises(ValueError):
                node3, create = self.usp.UspData.get_or_create(fpath,
                                                               use_first=False)

            # This should work now
            node4, create = self.usp.UspData.get_or_create(fpath,
                                                           use_first=True)
            self.assertFalse(create)
            self.assertIn(node4.pk, (node1.pk, node2.pk))

    def test_upload(self):
        self.upload_usp_family()

    def test_assign_from_structure(self):
        """
        Test using get_pseudos_from_structure
        """

        from aiida_castep.data import get_pseudos_from_structure
        from aiida.common import NotExistent

        self.upload_usp_family()
        STO = self.get_STO_structure()

        pseudo_list = get_pseudos_from_structure(STO, "STO")
        for kind in STO.kinds:
            self.assertIn(kind.name, pseudo_list)

        with self.assertRaises(NotExistent):
            STO.append_atom(symbols="Ba", position=(1, 1, 1))
            pseudo_list = get_pseudos_from_structure(STO, "STO")



"""
Utility module
"""
from __future__ import absolute_import
from collections import Counter
from aiida.common import InputValidationError


def get_castep_ion_line(name,
                        pos,
                        label=None,
                        spin=None,
                        occupation=None,
                        mix_num=None):
    """
    Generate a line in POSITIONS_ABS or POSITIONS_FRAC block

    :param name: A sting or tuple of the names
    :param pos: Position, a sequence of 3
    :param label: A string or tuple of the labels
    :param spin: Spins. (spin), (spin1, spin2), (x, y, z) or ((x1,y1,z1), (x2, y2, z2))
    :param occupation: tuple of two of the occupation number
    :param mix_num: mixture number, can be any integer but should not be repeated in a single cell file

    :return line: a string of the line to be added to the cell file
    """

    # Check if we are dealing with mixtures
    if isinstance(name, (tuple, list)):

        lines = [
            "{n:18} {x:18.10f} {y:18.10f} {z:18.10f}".format(n=n,
                                                             x=pos[0],
                                                             y=pos[1],
                                                             z=pos[2])
            for n in name
        ]

        assert sum(occupation) == 1, "Occupation need to sum up to 1"
        lines = [
            lines[i] + " MIXTURE= ({} {})".format(mix_num, occupation[i])
            for i in range(2)
        ]

        if label is not None:
            if not isinstance(label, (list, tuple)):
                label = [label, label]

            lines = [
                line + " LABEL={} ".format(l) for line, l in zip(lines, label)
            ]

        # Handle spin
        # spin might be (s1, s2) or (s1) or ((s11, s12, s12), (s21, s22, s23))
        if not isinstance(spin, (list, tuple)):
            spin = [spin, spin]
        elif len(spin) == 3:
            # Passed a 3D spin for both atoms
            spin = [spin, spin]

        if isinstance(spin[0], (list, tuple)):
            lines = [
                l + " SPIN=( {:.2f} {.2f} {.2f} )".format(*s)
                for l, s in zip(lines, spin)
            ]
        else:
            if None not in spin:
                lines = [l + " SPIN={}".format(s) for l, s in zip(lines, spin)]

        return "\n".join(lines)

    line = "{name:18} {x:18.10f} {y:18.10f} {z:18.10f}".format(name=name,
                                                               x=pos[0],
                                                               y=pos[1],
                                                               z=pos[2])

    if spin is not None:

        if isinstance(spin, (list, tuple)):
            line += " SPIN=({}, {}, {}) ".format(*spin)
        else:
            line += " SPIN={:.3f} ".format(spin)

    if label is not None:

        line += " LABEL={} ".format(label)

    return line


def _lowercase_dict(in_dict, dict_name):
    """
    Make sure the dictionary's keys are in lower case
    :param dict_name: A string of the name for the dictionary. Only used in
    warning message.
    """

    if isinstance(in_dict, dict):
        new_dict = dict(
            (str(key).lower(), value) for key, value in in_dict.items())
        if len(new_dict) != len(in_dict):
            num_items = Counter(str(key).lower() for key in in_dict.keys())
            double_keys = ",".join(
                [key for key, value in num_items if value > 1])
            raise InputValidationError(
                "Inside the dictionary '{}' there are the following keys that "
                "are repeated more than once when compared case-insensitively: {}."
                "This is not allowed.".format(dict_name, double_keys))
        return new_dict
    raise TypeError("_lowercase_dict accepts only dictionaries as argument")


def _uppercase_dict(in_dict, dict_name):
    """
    Make sure the dictionary's keys are in upper case
    :param dict_name: A string of the name for the dictionary. Only used in
    warning message.
    """
    if isinstance(in_dict, dict):
        new_dict = dict(
            (str(key).upper(), value) for key, value in in_dict.items())
        if len(new_dict) != len(in_dict):
            num_items = Counter(str(key).upper() for key in in_dict.keys())
            double_keys = ",".join(
                [key for key, value in num_items if value > 1])
            raise InputValidationError(
                "Inside the dictionary '{}' there are the following keys that "
                "are repeated more than once when compared case-insensitively: {}."
                "This is not allowed.".format(dict_name, double_keys))
        return new_dict
    raise TypeError("_uppercase_dict accepts only dictionaries as argument")



"""
Collection of fixtures for basic setup
"""
from __future__ import absolute_import
import pytest
import tempfile
import shutil


@pytest.fixture(scope='session')
def Path():
    """Load the Path class"""
    from pathlib import Path
    return Path


@pytest.fixture(scope='function')
def new_workdir(Path):
    """get a new temporary folder to use as the computer's workdir"""
    dirpath = tempfile.mkdtemp()
    yield Path(dirpath)
    shutil.rmtree(dirpath)


@pytest.fixture(scope='module')
def data_path(Path):
    """
    Return the directory to the data folder
    """
    this_file = Path(__file__)
    return (this_file.parent / 'data').resolve()


def test_data_path(data_path):
    """
    Test if the data_path exists
    """
    assert data_path.is_dir()



"""
This module contains the *base* workchain class which acts as the starting point of
implementing more complex ones
"""

import re
import numpy as np

from aiida.engine import WorkChain, if_, while_, ToContext, append_
from aiida.orm.nodes.data.base import to_aiida_type
from aiida.common import AttributeDict
import aiida.orm as orm
from aiida.plugins import DataFactory
from aiida.engine import CalcJob
from aiida_castep.data import get_pseudos_from_structure

from aiida_castep.common import (INPUT_LINKNAMES, OUTPUT_LINKNAMES)
from aiida_castep.calculations.helper import CastepHelper
from aiida_castep.calculations import CastepCalculation
from aiida_castep.calculations.tools import flat_input_param_validator
from .common import (UnexpectedCalculationFailure, register_error_handler,
                     ErrorHandlerReport)

KpointsData = DataFactory("array.kpoints")  # pylint: disable=invalid-name
StructureData = DataFactory("structure")  # pylint: disable=invalid-name
Dict = DataFactory("dict")  # pylint: disable=invalid-name

# pylint: disable=no-member

__version__ = '0.0.1'


class CastepBaseWorkChain(WorkChain):
    """
    A basic workchain for generic CASTEP calculations.
    We try to handle erros such as walltime exceeded or SCF not converged
    """

    _context_pain_dicts = ('parameters', 'settings')
    _calculation_class = CastepCalculation
    _verbose = False

    def __init__(self, *args, **kwargs):
        super(CastepBaseWorkChain, self).__init__(*args, **kwargs)

        if self._calculation_class is None or not issubclass(
                self._calculation_class, CalcJob):
            raise ValueError(
                'no valid CalcJob class defined for _calculation_class attribute'
            )

    @classmethod
    def define(cls, spec):
        """Define this workchain"""
        super(CastepBaseWorkChain, cls).define(spec)

        # The inputs
        spec.input('max_iterations',
                   valid_type=orm.Int,
                   default=lambda: orm.Int(10),
                   serializer=to_aiida_type,
                   help='Maximum number of restarts')
        spec.input(
            'reuse_folder',
            valid_type=orm.RemoteData,
            help=
            'Use a remote folder as the parent folder. Useful for restarts.',
            required=False)
        spec.input(
            'continuation_folder',
            valid_type=orm.RemoteData,
            help=
            'Use a remote folder as the parent folder. Useful for restarts.',
            required=False)
        spec.input('pseudos_family',
                   valid_type=orm.Str,
                   serializer=to_aiida_type,
                   required=False,
                   help='Pseudopotential family to be used')
        spec.input('kpoints_spacing',
                   valid_type=orm.Float,
                   required=False,
                   serializer=to_aiida_type,
                   help="Kpoint spacing")
        spec.input(
            'options',
            valid_type=orm.Dict,
            serializer=to_aiida_type,
            required=False,
            help=
            ('Options specifying resources, labels etc. Passed to the CalcJob.'
             'Avaliable options: queue_wallclock_limit, use_castep_bin'))
        spec.expose_inputs(cls._calculation_class, namespace='calc')
        spec.input('calc.parameters',
                   valid_type=orm.Dict,
                   serializer=to_aiida_type,
                   help='Input parameters, flat format is allowed.',
                   validator=flat_input_param_validator)

        spec.output('output_array', valid_type=orm.ArrayData, required=False)
        spec.output('output_trajectory',
                    valid_type=orm.ArrayData,
                    required=False)
        spec.output('output_bands', valid_type=orm.BandsData, required=True)
        spec.output('output_structure',
                    valid_type=orm.StructureData,
                    required=False)
        spec.output('output_parameters', valid_type=orm.Dict, required=True)
        spec.output('remote_folder', valid_type=orm.RemoteData)

        # Define the exit codes
        spec.exit_code(900, 'ERROR_INVALID_INPUTS', 'Input validate is failed')
        spec.exit_code(
            201, 'ERROR_TOTAL_WALLCLOCK_EXCEEDED',
            'The maximum length of the wallclocks has been exceeded')
        spec.exit_code(200, 'ERROR_MAXIMUM_ITERATIONS_EXCEEDED',
                       'The maximum number of iterations has been exceeded')
        spec.exit_code(301, 'ERROR_CASTEP_FAILURE',
                       'CASTEP generated error files and is not recoverable')
        spec.exit_code(302, 'ERROR_SCF_FAILURE',
                       'Cannot reach SCF convergence despite restart efforts')
        spec.exit_code(
            400, 'USER_REQUESTED_STOP',
            'The stop flag has been put in the .param file to request termination of the calculation.'
        )
        spec.exit_code(1000, 'UNKOWN_ERROR', 'Error is not kown')
        spec.exit_code(
            901, 'ERROR_ITERATION_RETURNED_NO_CALCULATION',
            'Completed one iteration but found not calculation returned')

        # Outline of the calculation
        spec.outline(
            cls.setup,
            cls.validate_inputs,
            if_(cls.should_dry_run)(
                cls.validate_dryrun_inputs,
                cls.run_dry_run,
                cls.inspect_dryrun,
            ),
            while_(cls.should_run_calculation)(
                cls.prepare_calculation,
                cls.run_calculation,
                cls.inspect_calculation,
            ),
            cls.results,
        )

    def setup(self):
        """Initialize internal variables"""
        self.ctx.calc_name = self._calculation_class.__name__
        self.ctx.stop_requested = None
        self.ctx.restart_calc = None
        self.ctx.restart_type = None
        self.ctx.is_finished = False
        self.ctx.unexpected_failure = False
        self.ctx.iteration = 0

    def validate_inputs(self):  # pylint: disable=too-many-branches, too-many-statements
        """Validate the inputs. Populate the inputs in the context
        This inputs is used as a staging area for the next calculation
        to be launched"""
        self.ctx.inputs = AttributeDict({
            'structure': self.inputs.calc.structure,
            'code': self.inputs.calc.code,
        })
        input_parameters = self.inputs.calc.parameters.get_dict()

        # Copy over the metadata
        self.ctx.inputs['metadata'] = AttributeDict(self.inputs.calc.metadata)
        self.ctx.inputs['metadata']['options'] = AttributeDict(
            self.inputs.calc.metadata.options)

        # propagate the settings to the inputs of the CalcJob
        if 'settings' in self.inputs.calc:
            self.ctx.inputs.settings = self.inputs.calc.settings.get_dict()
        else:
            self.ctx.inputs.settings = {}

        # Process the options in the input
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = {}
        self.ctx.options = options

        # Deal with the continuations
        use_bin = options.get('use_castep_bin', False)
        if use_bin:
            restart_suffix = 'castep_bin'
        else:
            restart_suffix = 'check'

        # Set the seed name
        seedname = self.inputs.calc.metadata.options.seedname

        # In case we are dealing with a plain inputs, extend any plain inputs
        helper = CastepHelper()
        param_dict = helper.check_dict(input_parameters)
        self.ctx.inputs['parameters'] = param_dict

        if self.inputs.get('continuation_folder'):
            self.ctx.inputs[INPUT_LINKNAMES[
                'parent_calc_folder']] = self.inputs.continuation_folder
            self.ctx.inputs.parameters['PARAM'][
                'continuation'] = 'parent/{}.{}'.format(
                    seedname, restart_suffix)
            self.ctx.inputs.parameters['PARAM'].pop('reuse', None)

        elif self.inputs.get('reuse_folder'):
            self.ctx.inputs[INPUT_LINKNAMES[
                'parent_calc_folder']] = self.inputs.reuse_folder
            self.ctx.inputs.parameters['PARAM'][
                'reuse'] = 'parent/{}.{}'.format(seedname, restart_suffix)
            self.ctx.inputs.parameters['PARAM'].pop('continuation', None)

        # Kpoints
        if self.inputs.calc.get('kpoints'):
            self.ctx.inputs.kpoints = self.inputs.calc.kpoints
        elif self.inputs.get('kpoints_spacing'):
            spacing = self.inputs.kpoints_spacing.value
            kpoints = KpointsData()
            # Here i set the cell directly
            # The set_cell_from_structure will consider the PBC
            # However for CASTEP a non-peroidic cell does not make any sense
            # So the default should be that the structure is peroidic
            kpoints.set_cell(self.inputs.calc.structure.cell)
            kpoints.set_kpoints_mesh_from_density(np.pi * 2 * spacing)
            self.ctx.inputs.kpoints = kpoints
        else:
            self.report('No valid kpoint input specified')
            return self.exit_codes.ERROR_INVALID_INPUTS
        # Pass extra kpoints
        exposed_inputs = self.exposed_inputs(CastepCalculation, 'calc')
        for key, value in exposed_inputs.items():
            if key.endswith('_kpoints'):
                self.ctx.inputs[key] = value

        # Validate the inputs related to pseudopotentials
        structure = self.inputs.calc.structure
        pseudos = self.inputs.calc.get('pseudos', None)
        pseudos_family = self.inputs.get('pseudos_family', None)
        if pseudos_family:
            pseudo_dict = get_pseudos_from_structure(structure,
                                                     pseudos_family.value)
            self.ctx.inputs.pseudos = pseudo_dict
        elif pseudos:
            self.ctx.inputs.pseudos = pseudos
        else:
            self.report('No valid pseudopotential input specified')
            return self.exit_codes.ERROR_INVALID_INPUTS
        return None

    def should_dry_run(self):  # pylint: disable=no-self-use
        """
        Do a dryrun to validate the inputs
        """
        return False  # We do not implement this for now
        #return 'do_dryrun' in self.inputs

    def validate_dryrun_inputs(self):
        pass

    def run_dry_run(self):
        pass

    def inspect_dryrun(self):
        pass

    def should_run_calculation(self):
        """Should we start the calculation (again)?"""
        return not self.ctx.is_finished and \
            self.ctx.iteration < self.inputs.max_iterations.value and \
            not self.ctx.stop_requested

    def prepare_calculation(self):
        """
        Prepare the inputs for the next calculation.
        """
        if self.ctx.restart_calc:
            # Different modes of restart
            if self.ctx.restart_mode == 'continuation':
                self.ctx.inputs.parameters['PARAM'][
                    'continuation'] = './parent/aiida.check'
                self.ctx.inputs.parameters['PARAM'].pop('reuse', None)
                self.ctx.inputs[INPUT_LINKNAMES[
                    'parent_calc_folder']] = self.ctx.restart_calc.outputs.remote_folder
            elif self.ctx.restart_type == 'reuse':
                self.ctx.inputs.parameters['PARAM'][
                    'reuse'] = './parent/aiida.check'
                self.ctx.inputs.parameters['PARAM'].pop('continuation', None)
                self.ctx.inputs[INPUT_LINKNAMES[
                    'parent_calc_folder']] = self.ctx.restart_calc.outputs.remote_folder
            else:
                self.ctx.inputs.parameters['PARAM'].pop('continuation', None)
                self.ctx.inputs.parameters['PARAM'].pop('reuse', None)
                self.ctx.inputs.pop(INPUT_LINKNAMES['parent_calc_folder'],
                                    None)

    def run_calculation(self):
        """
        Submit a new calculation, taking the input dictionary from the context at self.ctx.inputs
        """
        self.ctx.iteration += 1
        # Update the iterations in the inputs
        self.ctx.inputs.metadata['call_link_label'] = \
                                'iteration_{}'.format(self.ctx.iteration)

        try:
            unwrapped_inputs = AttributeDict(self.ctx.inputs)
        except AttributeError:
            raise ValueError(
                'no calculation input dictionary was defined in self.ctx.inputs'
            )

        inputs = self._prepare_process_inputs(unwrapped_inputs)
        calculation = self.submit(self._calculation_class, **inputs)

        self.report('launching {}<{}> iteration #{}'.format(
            self.ctx.calc_name, calculation.pk, self.ctx.iteration))

        return ToContext(calculations=append_(calculation))

    def inspect_calculation(self):
        """
        Analyse the results of the previous calculation, return/restart/abort if necessary
        """
        try:
            calculation = self.ctx.calculations[self.ctx.iteration - 1]
        except IndexError:
            self.report('iteration {} finished without returning a {}'.format(
                self.ctx.iteration, self.ctx.calc_name))
            return self.exit_codes.ERROR_ITERATION_RETURNED_NO_CALCULATION

        exit_code = None

        if calculation.is_finished_ok:
            self.report('{}<{}> completed successfully'.format(
                self.ctx.calc_name, calculation.pk))
            self.ctx.restart_calc = calculation
            self.ctx.is_finished = True

        # If the maximum number of iterations has been exceeded
        elif self.ctx.iteration >= self.inputs.max_iterations.value:
            self.report(
                'reached the maximumm number of iterations {}: last ran {}<{}>'
                .format(self.inputs.max_iterations.value, self.ctx.calc_name,
                        calculation.pk))
            exit_code = self.exit_codes.ERROR_MAXIMUM_ITERATIONS_EXCEEDED

        # Decide to retry or abort
        else:
            #exit_code = self._handle_calculation_sanity_checks(calculation)
            # Calculation failed, try to salvage it or handle any unexpected failures
            try:
                exit_code = self._handle_calculation_failure(calculation)
            except UnexpectedCalculationFailure as exception:
                exit_code = self._handle_unexpected_failure(
                    calculation, exception)
                self.ctx.unexpected_failure = True
        return exit_code

    def _prepare_process_inputs(self, inputs_dict):
        """Convert plain dictionary to Dict node"""
        out = AttributeDict(inputs_dict)
        for key in self._context_pain_dicts:
            if key in out:
                out[key] = Dict(dict=out[key])
        return out

    def _handle_calculation_failure(self, calculation):
        """Handle failure of calculation by refering to a range of handlers"""
        try:
            outputs = calculation.outputs[
                OUTPUT_LINKNAMES['results']].get_dict()
            _ = outputs['warnings']
            _ = outputs['parser_warnings']
        except (KeyError) as exception:
            raise UnexpectedCalculationFailure(exception)

        is_handled = False
        handler_report = None

        handlers = sorted(self._error_handlers,
                          key=lambda x: x.priority,
                          reverse=True)

        if not handlers:
            raise UnexpectedCalculationFailure(
                'no calculation error handlers were registered')

        for handler in handlers:
            handler_report = handler.method(self, calculation)

            if handler_report and handler_report.is_handled:
                is_handled = True

            if handler_report and handler_report.do_break:
                break

        # Raise error if not handled
        if not is_handled:
            raise UnexpectedCalculationFailure(
                'calculation failure was not handled')

        if handler_report:
            return handler_report.exit_code

        return None

    def results(self):
        """
        Attach the outputs specified in the output specification from the last completed calculation
        """
        self.report('workchain completed after {} iterations'.format(
            self.ctx.iteration))

        for name, port in self.spec().outputs.items():
            try:
                node = self.ctx.restart_calc.get_outgoing(
                    link_label_filter=name).one().node
            except ValueError:
                if port.required:
                    self.report(
                        "the process spec specifies the output '{}' as required but was not an output of {}<{}>"
                        .format(name, self.ctx.calc_name,
                                self.ctx.restart_calc.pk))
            else:
                self.out(name, node)
                if self._verbose:
                    self.report("attaching the node {}<{}> as '{}'".format(
                        node.__class__.__name__, node.pk, name))

    def _handle_unexpected_failure(self, calculation, exception=None):
        """
        The calculation has failed for an unknown reason and could not be handled.
        If the unexpected_failure flag is true, this is the second consecutive unexpected
        failure and we abort the workchain. Otherwise we restart once more.
        """
        if exception:
            self.report('{}'.format(exception))

        # if self.ctx.unexpected_failure:
        #     self.report(
        #         'failure of {}<{}> could not be handled for the second consecutive time'
        #         .format(self.ctx.calc_name, calculation.pk))
        #     return self.exit_codes.UNKOWN_ERROR

        # else:
        #     self.report(
        #         'failure of {}<{}> could not be handled, restarting once more'.
        #         format(self.ctx.calc_name, calculation.pk))

        self.report('failure of {}<{}> could not be handled'.format(
            self.ctx.calc_name, calculation.pk))
        return self.exit_codes.UNKOWN_ERROR


@register_error_handler(CastepBaseWorkChain, 900)
def _handle_scf_failure(self, calculation):
    """Handle case when SCF failed"""

    if 'ERROR_SCF_NOT_CONVERGED' in calculation.res.warnings:
        self.ctx.restart_calc = calculation
        self.ctx.restart_mode = None
        dot_castep = _get_castep_output_file(calculation)
        for idx, line in enumerate(dot_castep[:-50:-1]):
            model_match = re.match(r'Writing model to \w+\.(\w+)', line)
            # If the writing model is at the last line there is a good
            # Chance that it was interrupted
            if model_match and idx > 0 and model_match.group(1) == 'check':
                self.ctx.restart_mode = 'continuation'
                break

        param = self.ctx.inputs.parameters['PARAM']
        # Increase the SCF limit by 50%
        scf_limit = self.ctx.inputs.parameters['PARAM'].get(
            'max_scf_cycles', 30)
        scf_limit = int(scf_limit * 1.5)
        self.ctx.inputs.parameters['PARAM']['max_scf_cycles'] = scf_limit
        self.report('Increased SCF limit to: {}'.format(scf_limit))

        if param.get('metals_method') == 'edft' or param.get(
                'elec_method') == 'edft':
            return ErrorHandlerReport(True, True)

        # Reduce the mix charge amp
        mix_charge_amp = self.ctx.inputs.parameters['PARAM'].get(
            'mix_charge_amp', 0.8)
        if mix_charge_amp > 0.2:
            mix_charge_amp -= 0.1
        self.ctx.inputs.parameters['PARAM']['mix_charge_amp'] = mix_charge_amp

        # Reuce mix spin amp
        mix_spin_amp = self.ctx.inputs.parameters['PARAM'].get(
            'mix_spin_amp', 2)
        if mix_spin_amp > 0.5:
            mix_spin_amp -= 0.3
        self.ctx.inputs.parameters['PARAM']['mix_spin_amp'] = mix_spin_amp

        self.report(
            'Adjusted mix_charge_amp:{:.2f}, mix_spin_amp:{:.2f}'.format(
                mix_charge_amp, mix_spin_amp))

        return ErrorHandlerReport(True, True)
    return None


@register_error_handler(CastepBaseWorkChain, 500)
def _handle_walltime_limit(self, calculation):
    """Handle case when the walltime limit has reached"""

    if 'ERROR_TIMELIMIT_REACHED' in calculation.res.warnings:
        self.ctx.restart_calc = calculation
        self.ctx.restart_mode = None
        dot_castep = _get_castep_output_file(calculation)
        for nline, line in enumerate(dot_castep[::-1]):
            model_match = re.match(r'Writing model to \w+\.(\w+)', line)
            # If the writing model is at the last line there is a good
            # Chance that it was interrupted
            if model_match and nline > 0 and model_match.group(1) == 'check':
                self.ctx.restart_mode = 'continuation'
                self.report(
                    'dot castep indicate model has been written, trying continuation.'
                )
                break

        # If we are do not continue the run, try increase the wallclock

        if self.ctx.restart_mode is None:

            wclock = self.inputs.calc.metadata.options.get(
                'max_wallclock_seconds', 3600)
            wclock_limit = self.ctx.options.get('queue_wallclock_limit', None)
            if wclock_limit is None:
                pass
            elif wclock == wclock_limit:
                self.report('Cannot furhter increase the wallclock limit')
                return ErrorHandlerReport(False, True)
            elif wclock * 1.5 < wclock_limit:
                self.inputs.calc.metadata.options[
                    'max_wallclock_seconds'] = int(wclock * 1.5)
            else:
                self.inputs.calc.metadata.options[
                    'max_wallclock_seconds'] = int(wclock_limit)

            self.report('Adjusted the wallclock limit to {}'.format(
                self.inputs.calc.metadata.options['max_wallclock_seconds']))
            # Temporary fix - wait for next relax of aiida that allows customisation
            # of the valid cache for Process classes
            calculation.clear_hash()
            self.report('Cleared the hash of the failed calculation.')

        return ErrorHandlerReport(True, False)
    return None


@register_error_handler(CastepBaseWorkChain, 10000)
def _handle_stop_by_request(self, calculation):
    """Handle the case when the stop flag is raised by the user"""

    if 'ERROR_STOP_REQUESTED' in calculation.res.warnings:
        self.report('Stop is requested by user. Aborting the WorkChain.')
        self.ctx.restart_calc = calculation
        self.ctx.stop_requested = True
        calculation.clear_hash()
        self.report('Cleared the hash of the stopped calculation.')
        return ErrorHandlerReport(True, True,
                                  self.exit_codes.USER_REQUESTED_STOP)
    return None


def _get_castep_output_file(calculation):
    """Return a list of the lines in the retrieved dot castep file"""
    fname = calculation.get_option('output_filename')
    fcontent = calculation.outputs.retrieved.get_object_content(fname)
    return fcontent.split('\n')



"""
Command line interface for generating CASTEP help information
"""

from __future__ import print_function
from __future__ import absolute_import
import click
import sys
from aiida.cmdline.commands.cmd_data import verdi_data
from six.moves import map
from six.moves import zip


def dot_proc(iterable):
    """Provide a primitive progress bar"""
    print("Processing keywords")
    m = len(iterable) - 1
    print("=" * (m // 10))
    for n, i in enumerate(iterable):

        if n % 10 == 0:
            print(".", end="")
            sys.stdout.flush()
        if n == m:
            print("\n")
        yield i


def progress(func, *args, **kwargs):
    try:
        from tqdm import tqdm
    except ImportError:
        return dot_proc(func)
    else:
        return tqdm(func, *args, **kwargs)


@verdi_data.group('castep-help')
def helper_cmd():
    """Commandline interface for controlling helper information"""
    pass


@helper_cmd.command(name="generate")
@click.option("--castep-excutable",
              "-e",
              help="The excutable of CASTEP to be used",
              default="castep.serial")
@click.option("--save-as", "-s", help="override default path for saving file")
def generate(castep_excutable, save_as):
    """
    Generate help information file.

    The generated file will be saved as .castep_help_info_<version>.json
    at the $HOME by default.
    """
    from aiida_castep.calculations.helper.generate import (get_castep_commands,
                                                           parse_help_string)

    import subprocess as sbp
    import os
    try:
        castep_info = sbp.check_output([castep_excutable, "--version"],
                                       universal_newlines=True)
    except OSError:
        print("Not a valid CASTEP excutable. Aborted.")
        return

    version_num = None
    for line in castep_info.split("\n"):
        if "CASTEP version:" in line:
            version_num = line.split(":")[-1].strip()

    if save_as is None:
        fname = ".castep_help_info_{}.json".format(version_num)
        save_as = os.path.join(os.getenv("HOME"), fname)
    else:
        save_as = os.path.abspath(save_as)

    print("######## Version info of CASTEP ########")
    print(castep_info)
    print("Save path: {}".format(save_as))

    respond = click.prompt("Please check CASTEP version. [Y/N]")
    if not respond.lower() == "y":
        print("Aborted")
        return
    # Dictonary with short help lines
    all_keys = {}
    for key in ["basic", "inter", "expert"]:
        c, p = get_castep_commands(castep_excutable, key)
        all_keys.update(c)
        all_keys.update(p)

    # The full dictionary
    full_dict = {}

    for key in progress(all_keys):
        lines, k_type, k_level, v_type = parse_help_string(key)
        full_dict[key.lower()] = dict(help_short=all_keys[key],
                                      help_full="\n".join(lines),
                                      key_type=k_type,
                                      key_level=k_level,
                                      value_type=v_type)
    full_dict["_CASTEP_VERSION"] = version_num

    import json
    with open(save_as, "w") as json_out:
        json.dump(full_dict, json_out)
    print("Help information saved at {}".format(save_as))


@helper_cmd.command("show")
@click.argument("keyword")
def show_help(keyword):
    """
    Show help information given a keyword.
    Equivalent as castep -h <keyword> use the information previously saved.
    """
    helper = get_helper()
    h_text = helper.help_dict[keyword]["help_full"]
    print("")
    print(h_text)


@helper_cmd.command("list")
@click.option("--filter", "-f", help="Filter keys")
def list_keywords(filter):
    """
    Print out all avaliable keywords.
    Optional filter can be applied.
    """
    helper = get_helper()
    h_dict = helper.help_dict

    def print_keys(ktype):
        print("List of keywords in {} file:\n".format(ktype))
        for key, value in h_dict.items():

            if not isinstance(value, dict):
                continue
            if value["key_type"] == ktype:
                if filter is None or key.find(filter) != -1:
                    print("{:<40}{}".format(key, value["help_short"]))

        print("")

    print_keys("CELL")
    print_keys("PARAM")


@helper_cmd.command("listfile")
def list_file():
    """
    List files aviable to use.
    """

    from aiida_castep.calculations.helper import find_help_info
    pairs = find_help_info()
    if not pairs:
        print("No avaliale file detected")

    print("Avaliable files:")
    print("{:<67} | {:^10}".format('Path', 'version'))
    print("-" * 80)
    for path, version in pairs:
        if version == 0:
            version = 'NOT_SPECIFIED'
        print("{:<30} | {:>10}".format(path, version))


def get_helper(*args, **kwargs):
    """
    Get a helper object
    """
    from aiida_castep.calculations.helper import CastepHelper
    helper = CastepHelper(*args, **kwargs)
    return helper


if __name__ == "__main__":
    helper_cmd()



"""
Utility functions
"""
from __future__ import absolute_import
import os


def get_sto_structure():
    """Return a STO structure"""
    from aiida.plugins import DataFactory
    StructureData = DataFactory('structure')

    a = 3.905

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=["Sr"])
    s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
    s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
    s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"])
    s.append_atom(position=(0., a / 2, a / 2), symbols=["O"])
    s.label = "STO"
    return s


def get_sto_structure_with_tags():
    """Return a STO structure"""
    from aiida.plugins import DataFactory
    StructureData = DataFactory('structure')

    a = 3.905

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=["Sr"])
    s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
    s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
    s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"], name='O1')
    s.append_atom(position=(0., a / 2, a / 2), symbols=["O"], name='O2')
    s.label = "STO"
    return s


def get_mixture_cell():
    """Return a STO structure"""
    from aiida.plugins import DataFactory
    StructureData = DataFactory('structure')

    a = 3.905

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.),
                  symbols=["Sr", "Ti"],
                  weights=(0.5, 0.5),
                  name='SrTi')
    s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
    s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
    s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"], name='O1')
    s.append_atom(position=(0., a / 2, a / 2), symbols=["O"], name='O2')
    s.label = "STO"
    return s


def get_x2_structure(x):
    """Return a O2 molecule in a box"""
    from aiida.plugins import DataFactory
    StructureData = DataFactory('structure')
    a = 10

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=[x])
    s.append_atom(position=(1.4, 0., 0.), symbols=[x])
    s.label = "O2"
    return s


def get_data_abs_path():
    """Get the data folder path for the backend module"""
    test_moudule = os.path.split(os.path.abspath(__file__))[0]
    data_folder = os.path.join(test_moudule, "data")
    return data_folder
